from click import Argument, Option

from cloup.constraints.common import (
    format_param, format_param_list, get_param_label, join_with_and, param_value_is_set,
)
from tests.util import bool_opt, flag_opt, int_opt, parametrize, multi_opt, tuple_opt


@parametrize(
    'param_type,    value,      expected',
    (Argument,      None,       False),
    (Argument,      'bu',       True),
    (Option,        None,       False),
    (Option,        'bu',       True),
    (int_opt,       0,          True),
    (bool_opt,      False,      True),   # non-flag boolean opts are set even if False
    (flag_opt,      False,      False),
    (flag_opt,      True,       True),
    (tuple_opt,     (),         False),
    (tuple_opt,     (1, 2),     True),
    (multi_opt,     (),         False),
    (multi_opt,     (1, 2),     True),
    ids=lambda val: val.__name__ if callable(val) else None,
)
def test_param_value_is_set(param_type, value, expected):
    param = param_type(['-o'])
    actual = param_value_is_set(param, value)
    assert actual == expected


def test_get_param_label():
    assert get_param_label(Argument(['arg'])) == 'ARG'
    assert get_param_label(Option(['--opt'])) == '--opt'
    assert get_param_label(Option(['--opt', '-o'])) == '--opt'
    assert get_param_label(Option(['-o', '--opt'])) == '--opt'
    assert get_param_label(Option(['-o/-O', '--opt/--no-opt'])) == '--opt'


def test_format_param():
    assert format_param(Argument(['arg'])) == 'ARG'
    assert format_param(Option(['--opt'])) == '--opt'
    assert format_param(Option(['-o'])) == '-o'
    assert format_param(Option(['--opt', '-o'])) == '--opt (-o)'
    assert format_param(Option(['-o', '--opt'])) == '--opt (-o)'
    assert format_param(Option(['-o/-O', '--opt/--no-opt', 'blah'])) == '--opt (-o)'


def test_format_param_list():
    params = [
        Argument(['arg']),
        Option(['--one']),
        Option(['--two', '-t']),
    ]
    expected = (' ARG\n'
                ' --one\n'
                ' --two (-t)\n')
    assert format_param_list(params, indent=1) == expected


def test_join_with_and():
    assert join_with_and([]) == ''
    assert join_with_and('A') == 'A'
    assert join_with_and('ABC', sep='; ') == 'A; B and C'



"""
Example of option groups, "flat style".
"""
import click

import cloup
from cloup import OptionGroup, option
from cloup.constraints import If, RequireAtLeast, mutually_exclusive

_input = OptionGroup(
    'Input options',
    help='This is a very useful description of the group',
    constraint=mutually_exclusive,
)
_output = OptionGroup(
    'Output options',
    constraint=If('three', then=RequireAtLeast(1)),
)


@cloup.command('cloup_flat', align_option_groups=True, no_args_is_help=True)
# Input options
@_input.option('-o', '--one', help='1st input option')
@_input.option('--two', help='2nd input option')
@_input.option('--three', help='3rd input option')
# Output options
@_output.option('--four', help='1st output option')
@_output.option('--five', help='2nd output option')
@_output.option('--six', help='3rd output option')
# Other options
@option('--seven', help='first uncategorized option',
        type=click.Choice('yes no ask'.split()))
@option('--height', help='second uncategorized option')
def main(**kwargs):
    """ A CLI that does nothing. """
    print(kwargs)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

import os

import cloup


PROJ_DIR = os.path.abspath(os.path.join(__file__, '..', '..'))

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autodoc.typehints',
    'autoapi.extension',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'sphinx_panels',
    'sphinx_copybutton',  # adds a copy button to code blocks
    'versionwarning.extension',
    'sphinx_issues',  # link to GitHub issues and PRs
]

# General information about the project.
project = 'cloup'
copyright = "2020, Gianluca Gippetto"
author = "Gianluca Gippetto"
issues_github_path = "janLuke/cloup"

extlinks = {
    "issue": ("https://github.com/janLuke/cloup/issues/%s", "issue "),
    "pr": ("https://github.com/janLuke/cloup/pull/%s", "pull request "),
}

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|.
_version = tuple(map(str, cloup.__version_tuple__))
version = '.'.join(_version[:2])
release = '.'.join(_version[:3])

language = "en"

# Autodoc
autoclass_content = 'both'
autodoc_typehints = 'description'
set_type_checking_flag = True
typehints_fully_qualified = False

# Autoapi
autoapi_type = 'python'
autoapi_dirs = [os.path.join(PROJ_DIR, 'cloup')]
autoapi_template_dir = '_autoapi_templates'
templates_path = [autoapi_template_dir]
autoapi_keep_files = True
autoapi_add_toctree_entry = False
autoapi_python_class_content = 'both'
autoapi_options = [
    'members',
    'undoc-members',
    'show-inheritance',
    'show-module-summary',
    'special-members',
    'imported-members',
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/', None),
    'Click': ('https://click.palletsprojects.com', None)
}

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = '.rst'
master_doc = 'index'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output ---------------------------------------------------
html_title = f"cloup v{version}"
html_theme = "furo"
html_theme_options = {
    "light_logo": "logo.svg",
    "dark_logo": "logo-dark-mode.svg",
    "sidebar_hide_name": True,
}
html_css_files = [
    'styles/extensions-overrides.css',
    'styles/theme-overrides.css',
]
pygments_style = 'default'  # name of the Pygments (syntax highlighting) style

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

primary_color = "#0094ff"
darker_primary_color = "#007bd3"
def primary_color_alpha(alpha):
    return "#2a5adf" + "{:02x}".format(int(alpha * 255))

panels_css_variables = {
    "tabs-color-label-active": darker_primary_color,
    "tabs-color-label-inactive": "var(--color-foreground-muted)",
    "tabs-color-overline": "var(--tabs--border)",
    "tabs-color-underline": "var(--tabs--border)",
}
panels_add_bootstrap_css = False


# -- Version warning -----------------------------------------------------------
versionwarning_messages = {
    "latest": (
        'This is the documentation for the main development branch of Cloup. '
        'The documentation for the latest stable release is '
        '<a href="/en/stable/">here</a>.'
    ),
}
# versionwarning_project_version = "latest"    # For debugging locally
versionwarning_body_selector = 'article[role="main"]'

# Whether to render to-do notes.
todo_include_todos = False

# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'cloupdoc'

# -- Options for LaTeX output ------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'cloup.tex',
     'cloup Documentation',
     'Gianluca Gippetto', 'manual'),
]



import os
import shutil
from glob import glob
import argparse
from itertools import chain

parser = argparse.ArgumentParser(
    description="""
    Remove full directories and files matching the provided glob patterns.

    If -r/--recursive, the pattern '**' will match any files and zero or more
    directories and subdirectories.
    """
)
parser.add_argument('paths', nargs='+')
parser.add_argument('-r', '--recursive', action='store_true',
                    help="Use recursive globs, i.e. the pattern '**' will match any "
                         "files and zero or more directories and subdirectories.")
parser.add_argument('-d', '--dry-run', action='store_true',
                    help='Do not remove files, just print a list of them')

args = parser.parse_args()

paths = set(chain.from_iterable(
    glob(arg, recursive=args.recursive)
    for arg in args.paths
))

if args.dry_run:
    print('\n'.join(paths))
else:
    for path in paths:
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)
        print('Removed', path)



from functools import partial

from click.testing import CliRunner
from pytest import fixture

from tests.example_command import make_example_command
from tests.example_group import make_example_group


@fixture()
def runner():
    runner = CliRunner()
    runner.invoke = partial(runner.invoke, catch_exceptions=False)
    return runner


@fixture(scope='session')
def get_example_command():
    def get_command(tabular_help=True, align_option_groups=True):
        return make_example_command(
            align_option_groups=align_option_groups, tabular_help=tabular_help)

    return get_command


@fixture(scope='session')
def get_example_group():
    def get_group(align_sections):
        return make_example_group(align_sections=align_sections)

    return get_group



import re

import click
import pytest

import cloup
from cloup._util import reindent
from tests.util import new_dummy_func


def test_command_handling_of_unknown_argument():
    with pytest.raises(TypeError, match='HINT: you set cls='):
        cloup.command(cls=click.Command, align_option_groups=True)(new_dummy_func())
    with pytest.raises(TypeError, match='nonexisting') as info:
        cloup.command(nonexisting=True)(new_dummy_func())
    assert re.search(str(info.value), 'HINT') is None


def test_group_raises_if_cls_is_not_subclass_of_click_Group():
    cloup.group()
    cloup.group(cls=click.Group)
    cloup.group(cls=cloup.Group)
    with pytest.raises(TypeError):
        cloup.group(cls=click.Command)


def test_group_handling_of_unknown_argument():
    with pytest.raises(TypeError, match='HINT'):
        cloup.group(cls=click.Group, align_sections=True)(new_dummy_func())
    with pytest.raises(TypeError) as info:
        cloup.group(unexisting_arg=True)(new_dummy_func())
    assert re.search(str(info.value), 'HINT') is None


def test_command_works_with_no_parameters(runner):
    cmd = cloup.Command(name='cmd', callback=new_dummy_func())
    res = runner.invoke(cmd, '--help')
    assert res.output == reindent("""
        Usage: cmd [OPTIONS]

        Options:
          --help  Show this message and exit.
    """)


def test_group_works_with_no_params_and_subcommands(runner):
    cmd = cloup.Group(name='cmd')
    res = runner.invoke(cmd, '--help')
    assert res.output == reindent("""
        Usage: cmd [OPTIONS] COMMAND [ARGS]...

        Options:
          --help  Show this message and exit.
    """)


class TestDidYouMean:
    @pytest.fixture(scope="class")
    def cmd(self):
        cmd = cloup.Group(name="cmd")
        subcommands = [
            ('install', ['ins']),
            ('remove', ['rm']),
            ('clear', [])
        ]
        for name, aliases in subcommands:
            cmd.add_command(
                cloup.Command(name=name, aliases=aliases, callback=new_dummy_func()))
        return cmd

    def test_with_no_matches(self, runner, cmd):
        res = runner.invoke(cmd, 'asdfdsgdfgdf')
        assert res.output == reindent("""
            Usage: cmd [OPTIONS] COMMAND [ARGS]...
            Try 'cmd --help' for help.

            Error: No such command 'asdfdsgdfgdf'.
        """)

    def test_with_one_match(self, runner, cmd):
        res = runner.invoke(cmd, 'clearr')
        assert res.output == reindent("""
            Usage: cmd [OPTIONS] COMMAND [ARGS]...
            Try 'cmd --help' for help.

            Error: No such command 'clearr'. Did you mean 'clear'?
        """)

    def test_with_multiple_matches(self, runner, cmd):
        res = runner.invoke(cmd, 'inst')
        assert res.output == reindent("""
            Usage: cmd [OPTIONS] COMMAND [ARGS]...
            Try 'cmd --help' for help.

            Error: No such command 'inst'. Did you mean one of these?
               ins
               install
        """)



"""Tests for the "option groups" feature/module."""
from textwrap import dedent
from typing import cast

import click
import pytest
from click import pass_context

import cloup
from cloup import OptionGroup, option, option_group
from cloup._util import pick_non_missing
from cloup.typing import MISSING
from cloup.constraints import RequireAtLeast, mutually_exclusive
from tests.util import (make_options, new_dummy_func, parametrize, pick_first_bool)


def test_error_message_if_first_arg_is_not_a_string():
    with pytest.raises(
        TypeError, match="the first argument of @option_group must be its title"
    ):
        @option_group(
            option('--one'),
            option('--two'),
        )
        def f(one, two):
            pass


@parametrize(
    ['tabular_help', 'align_option_groups'],
    pytest.param(True, True, id='tabular-aligned'),
    pytest.param(True, False, id='tabular-non_aligned'),
    pytest.param(False, None, id='linear'),
)
def test_option_groups_are_correctly_displayed_in_help(
    runner, tabular_help, align_option_groups, get_example_command
):
    cmd = get_example_command(
        tabular_help=tabular_help,
        align_option_groups=align_option_groups
    )
    result = runner.invoke(cmd, args=('--help',))
    assert result.exit_code == 0
    assert result.output.strip() == cmd.expected_help


def test_option_group_constraints_are_checked(runner, get_example_command):
    cmd = get_example_command(align_option_groups=False)

    result = runner.invoke(cmd, args='arg1 --one=1')
    assert result.exit_code == 0

    result = runner.invoke(cmd, args='arg1 --one=1 --three=3 --five=4')
    assert result.exit_code == 0

    result = runner.invoke(cmd, args='arg1 --one=1 --three=3')
    assert result.exit_code == 2
    error_prefix = ('Error: when --three is set, at least 1 of the following '
                    'parameters must be set')
    assert error_prefix in result.output


def test_option_group_decorator_raises_if_group_is_passed_to_contained_option():
    with pytest.raises(ValueError):
        @option_group(
            'a group',
            cloup.option('--opt', group=OptionGroup('another group'))
        )
        def f(opt):
            pass


def test_option_group_decorator_raises_for_no_options():
    with pytest.raises(ValueError):
        cloup.option_group('grp')


@pytest.mark.parametrize(
    'cmd_value', [MISSING, None, True, False],
    ids=lambda val: f'cmd_{val}'
)
@pytest.mark.parametrize(
    'ctx_value', [MISSING, None, True, False],
    ids=lambda val: f'ctx_{val}'
)
def test_align_option_groups_context_setting(runner, ctx_value, cmd_value):
    should_align = pick_first_bool([cmd_value, ctx_value], default=True)
    cxt_settings = pick_non_missing(dict(
        align_option_groups=ctx_value,
        terminal_width=80,
    ))
    cmd_kwargs = pick_non_missing(dict(
        align_option_groups=cmd_value,
        context_settings=cxt_settings
    ))

    @cloup.command(**cmd_kwargs)
    @cloup.option_group('First group', option('--opt', help='first option'))
    @cloup.option_group('Second group', option('--much-longer-opt', help='second option'))
    @pass_context
    def cmd(ctx, one, much_longer_opt):
        assert cmd.must_align_groups(ctx) == should_align

    result = runner.invoke(cmd, args=('--help',))
    start = result.output.find('First')
    if should_align:
        expected = """
            First group:
              --opt TEXT              first option

            Second group:
              --much-longer-opt TEXT  second option

            Other options:
              --help                  Show this message and exit."""
    else:
        expected = """
            First group:
              --opt TEXT  first option

            Second group:
              --much-longer-opt TEXT  second option

            Other options:
              --help  Show this message and exit."""

    expected = dedent(expected).strip()
    end = start + len(expected)
    assert result.output[start:end] == expected


def test_context_settings_propagate_to_children(runner):
    @cloup.group(context_settings=dict(align_option_groups=False))
    def grp():
        pass

    @grp.command()
    @pass_context
    def cmd(ctx):
        assert cmd.must_align_option_groups(ctx) is False

    runner.invoke(grp, ('cmd',))


def test_that_neither_optgroup_nor_its_options_are_shown_if_optgroup_is_hidden(runner):
    @cloup.command('name')
    @cloup.option_group(
        'Hidden group',
        cloup.option('--one'),
        hidden=True
    )
    def cmd():
        pass

    result = runner.invoke(cmd, args=('--help',), catch_exceptions=False)
    assert 'Hidden group' not in result.output
    assert '--one' not in result.output


def test_that_optgroup_is_hidden_if_all_its_options_are_hidden(runner):
    @cloup.command('name')
    @cloup.option_group(
        'Hidden group',
        cloup.option('--one', hidden=True),
        cloup.option('--two', hidden=True),
    )
    def cmd():
        pass

    assert cmd.option_groups[0].hidden
    result = runner.invoke(cmd, args=('--help',), catch_exceptions=False)
    assert 'Hidden group' not in result.output


def test_option_group_options_setter_set_the_hidden_attr_of_options():
    opts = make_options('abc')
    group = OptionGroup('name')
    group.options = opts
    assert not any(opt.hidden for opt in opts)
    group.hidden = True
    group.options = opts
    assert all(opt.hidden for opt in opts)


def test_option_group_with_constrained_subgroups(runner):
    @cloup.command()
    @option_group(
        "Some options",
        RequireAtLeast(1)(
            option('-a', is_flag=True),
            option('-b', is_flag=True)
        ),
        mutually_exclusive(
            option('-c', is_flag=True),
            option('-d', is_flag=True),
        ),
        option('-e', is_flag=True),
    )
    def cmd(a, b, c, d, e):
        pass

    cmd = cast(cloup.Command, cmd)
    assert len(cmd.option_groups) == 1
    assert len(cmd.option_groups[0]) == 5

    assert runner.invoke(cmd, ['-abc']).exit_code == 0
    assert 'Error: at least 1' in runner.invoke(cmd, ['-c']).output
    assert 'mutually exclusive' in runner.invoke(cmd, ['-acd']).output

    expected_help = dedent("""
        Usage: cmd [OPTIONS]

        Some options:
          -a
          -b
          -c
          -d
          -e

        Other options:
          --help  Show this message and exit.
    """).lstrip()
    actual_help = runner.invoke(cmd, ['--help']).output
    assert actual_help == expected_help


def test_usage_of_constraints_as_decorators_inside_option_group(runner):
    @cloup.command()
    @cloup.option_group(
        "Options",
        mutually_exclusive(
            cloup.option('-a', is_flag=True),
            cloup.option('-b', is_flag=True),
            cloup.option('-c', is_flag=True),
        ),
        cloup.option('-d', is_flag=True),
    )
    def cmd(a, b, c, d):
        pass

    expected_help = dedent("""
        Usage: cmd [OPTIONS]

        Options:
          -a
          -b
          -c
          -d

        Other options:
          --help  Show this message and exit.
    """).lstrip()

    res = runner.invoke(cmd, args=['--help'])
    assert res.output == expected_help

    assert runner.invoke(cmd, args=[]).exit_code == 0
    assert runner.invoke(cmd, args='-d'.split()).exit_code == 0
    assert runner.invoke(cmd, args='-ad'.split()).exit_code == 0

    res = runner.invoke(cmd, args='-ab'.split())
    assert res.exit_code == click.UsageError.exit_code
    assert 'mutually exclusive' in res.output


def test_option_groups_raises_if_input_decorator_add_an_argument():
    with pytest.raises(
        TypeError, match='only parameter of type `Option` can be added to option groups'
    ):
        option_group(
            'Title',
            mutually_exclusive(
                cloup.argument('arg'),
                cloup.option('--opt')
            )
        )(new_dummy_func())



from pprint import pprint

import cloup
from cloup import argument, command, option, option_group
from cloup.constraints import ErrorFmt, mutually_exclusive


@command(aliases=["r", "re"])
@argument("script_path", help="Script path.", type=cloup.Path(), required=True)
@argument("scene_names", help="Name of the scenes.", required=False, nargs=-1)
@option_group(
    "Global options",
    option(
        "-c", "--config_file",
        help="Specify the configuration file to use for render settings.",
    ),
    option(
        "--custom_folders", is_flag=True,
        help="Use the folders defined in the [custom_folders] section of the "
             "config file to define the output folder structure.",
    ),
    option(
        "--disable_caching", is_flag=True,
        help="Disable the use of the cache (still generates cache files).",
    ),
    option("--flush_cache", is_flag=True, help="Remove cached partial movie files."),
    option("--tex_template", help="Specify a custom TeX template file."),
    option(
        "-v", "--verbosity",
        type=cloup.Choice(
            ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
            case_sensitive=False,
        ),
        help="Verbosity of CLI output. Changes ffmpeg log level unless 5+.",
    ),
    option(
        "--notify_outdated_version/--silent", is_flag=True, default=None,
        help="Display warnings for outdated installation.",
    ),
)
@option_group(
    "Output options",
    option(
        "-o", "--output_file", multiple=True,
        help="Specify the filename(s) of the rendered scene(s).",
    ),
    option(
        "--write_to_movie", is_flag=True, default=None,
        help="Write to a file.",
    ),
    option(
        "--media_dir", type=cloup.Path(),
        help="Path to store rendered videos and latex.",
    ),
    option(
        "--log_dir", type=cloup.Path(), help="Path to store render logs."
    ),
    option(
        "--log_to_file", is_flag=True,
        help="Log terminal output to file.",
    ),
)
@option_group(
    "Render Options",
    option(
        "-n", "--from_animation_number",
        help="Start rendering from n_0 until n_1. If n_1 is left unspecified, "
             "renders all scenes after n_0.",
    ),
    option(
        "-a", "--write_all", is_flag=True,
        help="Render all scenes in the input file.",
    ),
    option(
        "-f", "--format", "file_format", default="mp4",
        type=cloup.Choice(["png", "gif", "mp4"], case_sensitive=False),
    ),
    option("-s", "--save_last_frame", is_flag=True),
    option(
        "-q", "--quality", default="h",
        type=cloup.Choice(["l", "m", "h", "p", "k"], case_sensitive=False),
        help="""\b
            Resolution and framerate of the render:
            l = 854x480  30FPS,   m = 1280x720 30FPS,
            h = 1920x1080 60FPS,  p = 2560x1440 60FPS,
            k = 3840x2160 60FPS
            """,
    ),
    option(
        "-r", "--resolution",
        help="Resolution in (W,H) for when 16:9 aspect ratio isn't possible.",
    ),
    option(
        "--fps", "--frame_rate", "frame_rate", type=float,
        help="Render at this frame rate.",
    ),
    mutually_exclusive.rephrased(
        error=f"{ErrorFmt.error}\n"
              f"Use --renderer, the other two options are deprecated."
    )(
        option(
            "--renderer",
            type=cloup.Choice(["cairo", "opengl", "webgl"], case_sensitive=False),
            help="Select a renderer for your Scene.",
        ),
        option(
            "--use_opengl_renderer", is_flag=True, hidden=True,
            help="(Deprecated) Use --renderer=opengl.",
        ),
        option(
            "--use_cairo_renderer", is_flag=True, hidden=True,
            help="(Deprecated) Use --renderer=cairo.",
        ),
    ),
    option(
        "--webgl_renderer_path", type=cloup.Path(),
        help="The path to the WebGL frontend.",
    ),
    option(
        "-t", "--transparent", is_flag=True,
        help="Render scenes with alpha channel."
    ),
)
@option_group(
    "Ease of access options",
    option(
        "--progress_bar", default="display",
        type=cloup.Choice(["display", "leave", "none"], case_sensitive=False),
        help="Display progress bars and/or keep them displayed.",
    ),
    option(
        "-p", "--preview", is_flag=True,
        help="Preview the Scene's animation. OpenGL does a live preview in a "
             "popup window. Cairo opens the rendered video file in the system "
             "default media player.",
    ),
    option(
        "-f", "--show_in_file_browser", is_flag=True,
        help="Show the output file in the file browser.",
    ),
    option("--jupyter", is_flag=True, help="Using jupyter notebook magic."),
)
def render(**kwargs):
    """Render some or all scenes defined in a Python script."""
    pprint(kwargs, indent=2)



from typing import cast
from unittest.mock import Mock

import click
from click import Context
from pytest import fixture

import cloup
from cloup import Command


@fixture()
def dummy_ctx():
    """Useful for testing methods that needs a Context but don't actually use it."""
    dummy = Mock(spec_set=Context(Command('name', params=[])))
    dummy.command.__class__ = Command
    return dummy


@fixture()
def sample_cmd() -> Command:
    """Useful for testing constraints against a variety of parameter kinds.
    Parameters have names that should make easy to remember their "kind"
    without the need for looking up this code."""
    @cloup.command()
    # Optional arguments
    @click.argument('arg1', required=False)
    @click.argument('arg2', required=False)
    # Plain options without default
    @cloup.option('--str-opt')
    @cloup.option('--int-opt', type=int)
    @cloup.option('--bool-opt', type=bool)
    # Flags
    @cloup.option('--flag / --no-flag')
    @cloup.option('--flag2', is_flag=True)
    # Options with default
    @cloup.option('--def1', default=1)
    @cloup.option('--def2', default=2)
    # Options that take a tuple
    @cloup.option('--tuple', nargs=2, type=int)
    # Options that can be specified multiple times
    @cloup.option('--mul1', type=int, multiple=True)
    @cloup.option('--mul2', type=int, multiple=True)
    def f(**kwargs):
        print('It works')

    return cast(Command, f)



import warnings
from typing import Any, Callable, Dict, List, Optional, Type

import click

import cloup
from cloup._util import coalesce, pick_non_missing
from cloup.formatting import HelpFormatter
from cloup.typing import MISSING, Possibly


def _warn_if_formatter_settings_conflict(
    ctx_key: str, formatter_key: str, ctx_kwargs: dict, formatter_settings: dict
) -> None:
    if ctx_kwargs.get(ctx_key) and formatter_settings.get(formatter_key):
        from textwrap import dedent
        formatter_arg = f'formatter_settings.{formatter_key}'
        warnings.warn(dedent(f"""
        You provided both {ctx_key} and {formatter_arg} as arguments of a Context.
        Unless you have a particular reason, you should set only one of them..
        If you use both, {formatter_arg} will be used by the formatter.
        You can suppress this warning by setting:

            cloup.warnings.formatter_settings_conflict = False
        """))


class Context(click.Context):
    """A custom context for Cloup.

    Look up :class:`click.Context` for the list of all arguments.

    .. versionadded:: 0.9.0
        added the ``check_constraints_consistency`` parameter.

    .. versionadded:: 0.8.0

    :param ctx_args:
        arguments forwarded to :class:`click.Context`.
    :param align_option_groups:
        if True, align the definition lists of all option groups of a command.
        You can override this by setting the corresponding argument of ``Command``
        (but you probably shouldn't: be consistent).
    :param align_sections:
        if True, align the definition lists of all subcommands of a group.
        You can override this by setting the corresponding argument of ``Group``
        (but you probably shouldn't: be consistent).
    :param show_subcommand_aliases:
        whether to show the aliases of subcommands in the help of a ``cloup.Group``.
    :param show_constraints:
        whether to include a "Constraint" section in the command help (if at
        least one constraint is defined).
    :param check_constraints_consistency:
        enable additional checks for constraints which detects mistakes of the
        developer (see :meth:`cloup.Constraint.check_consistency`).
    :param formatter_settings:
        keyword arguments forwarded to :class:`HelpFormatter` in ``make_formatter``.
        This args are merged with those of the (eventual) parent context and then
        merged again (being overridden) by those of the command.
        **Tip**: use the static method :meth:`HelpFormatter.settings` to create this
        dictionary, so that you can be guided by your IDE.
    :param ctx_kwargs:
        keyword arguments forwarded to :class:`click.Context`.
    """
    formatter_class: Type[HelpFormatter] = HelpFormatter

    def __init__(
        self, *ctx_args,
        align_option_groups: Optional[bool] = None,
        align_sections: Optional[bool] = None,
        show_subcommand_aliases: Optional[bool] = None,
        show_constraints: Optional[bool] = None,
        check_constraints_consistency: Optional[bool] = None,
        formatter_settings: Dict[str, Any] = {},
        **ctx_kwargs,
    ):
        super().__init__(*ctx_args, **ctx_kwargs)
        self.align_option_groups = coalesce(
            align_option_groups,
            getattr(self.parent, 'align_option_groups', None),
        )
        self.align_sections = coalesce(
            align_sections,
            getattr(self.parent, 'align_sections', None),
        )
        self.show_subcommand_aliases = coalesce(
            show_subcommand_aliases,
            getattr(self.parent, 'show_subcommand_aliases', None),
        )
        self.show_constraints = coalesce(
            show_constraints,
            getattr(self.parent, 'show_constraints', None),
        )
        self.check_constraints_consistency = coalesce(
            check_constraints_consistency,
            getattr(self.parent, 'check_constraints_consistency', None)
        )

        if cloup.warnings.formatter_settings_conflict:
            _warn_if_formatter_settings_conflict(
                'terminal_width', 'width', ctx_kwargs, formatter_settings)
            _warn_if_formatter_settings_conflict(
                'max_content_width', 'max_width', ctx_kwargs, formatter_settings)

        #: Keyword arguments for the HelpFormatter. Obtained by merging the options
        #: of the parent context with the one passed to this context. Before creating
        #: the help formatter, these options are merged with the (eventual) options
        #: provided to the command (having higher priority).
        self.formatter_settings = {
            **getattr(self.parent, 'formatter_settings', {}),
            **formatter_settings,
        }

    def get_formatter_settings(self) -> Dict[str, Any]:
        return {
            'width': self.terminal_width,
            'max_width': self.max_content_width,
            **self.formatter_settings,
            **getattr(self.command, 'formatter_settings', {})
        }

    def make_formatter(self) -> HelpFormatter:
        opts = self.get_formatter_settings()
        return self.formatter_class(**opts)

    @staticmethod
    def settings(
        *, auto_envvar_prefix: Possibly[bool] = MISSING,
        default_map: Possibly[Dict[str, Any]] = MISSING,
        terminal_width: Possibly[int] = MISSING,
        max_content_width: Possibly[int] = MISSING,
        resilient_parsing: Possibly[bool] = MISSING,
        allow_extra_args: Possibly[bool] = MISSING,
        allow_interspersed_args: Possibly[bool] = MISSING,
        ignore_unknown_options: Possibly[bool] = MISSING,
        help_option_names: Possibly[List[str]] = MISSING,
        token_normalize_func: Possibly[Callable[[str], str]] = MISSING,
        color: Possibly[bool] = MISSING,
        show_default: Possibly[bool] = MISSING,
        align_option_groups: Possibly[bool] = MISSING,
        align_sections: Possibly[bool] = MISSING,
        show_subcommand_aliases: Possibly[bool] = MISSING,
        show_constraints: Possibly[bool] = MISSING,
        check_constraints_consistency: Possibly[bool] = MISSING,
        formatter_settings: Possibly[Dict[str, Any]] = MISSING,
    ) -> Dict[str, Any]:
        """Utility method for creating a ``context_settings`` dictionary.

        :param auto_envvar_prefix:
            the prefix to use for automatic environment variables. If this is
            `None` then reading from environment variables is disabled. This
            does not affect manually set environment variables which are always
            read.
        :param default_map:
            a dictionary (like object) with default values for parameters.
        :param terminal_width:
            the width of the terminal. The default is inherited from parent
            context. If no context defines the terminal width then auto-detection
            will be applied.
        :param max_content_width:
            the maximum width for content rendered by Click (this currently
            only affects help pages). This defaults to 80 characters if not
            overridden. In other words: even if the terminal is larger than
            that, Click will not format things wider than 80 characters by
            default.  In addition to that, formatters might add some safety
            mapping on the right.
        :param resilient_parsing:
            if this flag is enabled then Click will parse without any
            interactivity or callback invocation. Default values will also be
            ignored. This is useful for implementing things such as completion
            support.
        :param allow_extra_args:
            if this is set to `True` then extra arguments at the end will not
            raise an error and will be kept on the context. The default is to
            inherit from the command.
        :param allow_interspersed_args:
            if this is set to `False` then options and arguments cannot be
            mixed.  The default is to inherit from the command.
        :param ignore_unknown_options:
            instructs click to ignore options it does not know and keeps them
            for later processing.
        :param help_option_names:
            optionally a list of strings that define how the default help
            parameter is named. The default is ``['--help']``.
        :param token_normalize_func:
            an optional function that is used to normalize tokens (options,
            choices, etc.). This for instance can be used to implement
            case-insensitive behavior.
        :param color:
            controls if the terminal supports ANSI colors or not. The default
            is auto-detection. This is only needed if ANSI codes are used in
            texts that Click prints which is by default not the case. This for
            instance would affect help output.
        :param show_default: Show defaults for all options. If not set,
            defaults to the value from a parent context. Overrides an
            option's ``show_default`` argument.
        :param align_option_groups:
            if True, align the definition lists of all option groups of a command.
            You can override this by setting the corresponding argument of ``Command``
            (but you probably shouldn't: be consistent).
        :param align_sections:
            if True, align the definition lists of all subcommands of a group.
            You can override this by setting the corresponding argument of ``Group``
            (but you probably shouldn't: be consistent).
        :param show_subcommand_aliases:
            whether to show the aliases of subcommands in the help of a ``cloup.Group``.
        :param show_constraints:
            whether to include a "Constraint" section in the command help (if at
            least one constraint is defined).
        :param check_constraints_consistency:
            enable additional checks for constraints which detects mistakes of the
            developer (see :meth:`cloup.Constraint.check_consistency`).
        :param formatter_settings:
            keyword arguments forwarded to :class:`HelpFormatter` in ``make_formatter``.
            This args are merged with those of the (eventual) parent context and then
            merged again (being overridden) by those of the command.
            **Tip**: use the static method :meth:`HelpFormatter.settings` to create this
            dictionary, so that you can be guided by your IDE.
        """
        return pick_non_missing(locals())



import dataclasses as dc
import inspect
import shutil
import textwrap
from itertools import chain
from typing import (
    Any, Callable, Dict, Iterable, Iterator, Optional, Sequence, TYPE_CHECKING,
    Tuple, Union, cast,
)

from cloup._util import click_version_ge_8_1
from cloup.formatting._util import unstyled_len

if TYPE_CHECKING:
    from .sep import RowSepPolicy, SepGenerator

import click
from click.formatting import wrap_text

from cloup._util import (
    check_positive_int, identity, indent_lines, make_repr,
    pick_non_missing
)
from ..typing import MISSING, Possibly
from cloup.styling import HelpTheme, IStyle

Definition = Tuple[str, Union[str, Callable[[int], str]]]


@dc.dataclass()
class HelpSection:
    """A container for a help section data."""
    heading: str
    """Help section title."""

    definitions: Sequence[Definition]
    """Rows with 2 columns each. The 2nd element of each row can also be a function
    taking an integer (the available width for the 2nd column) and returning a string."""

    help: Optional[str] = None
    """(Optional) long description of the section."""

    constraint: Optional[str] = None
    """(Optional) option group constraint description."""


# noinspection PyMethodMayBeStatic
class HelpFormatter(click.HelpFormatter):
    """
    A custom help formatter. Features include:

    - more attributes for controlling the output of the formatter
    - a ``col1_width`` parameter in :meth:`write_dl` that allows Cloup to align
      multiple definition lists without resorting to hacks
    - a "linear layout" for definition lists that kicks in when the available
      terminal width is not enough for the standard 2-column layout
      (see argument ``col2_min_width``)
    - the first column width, when not explicitly given in ``write_dl`` is
      computed excluding the rows that exceed ``col1_max_width``
      (called ``col_max`` in ``write_dl`` for compatibility with Click).

    .. versionchanged:: 0.9.0
        the ``row_sep`` parameter now:

        - is set to ``None`` by default and ``row_sep=""`` corresponds to an
          empty line between rows
        - must not ends with ``\\n``; the formatter writes a newline just after
          it (when it's not ``None``), so a newline at the end is always enforced
        - accepts instances of :class:`~cloup.formatting.sep.SepGenerator` and
          :class:`~cloup.formatting.sep.RowSepPolicy`.

    .. versionadded:: 0.8.0

    :param indent_increment:
        width of each indentation increment.
    :param width:
        content line width, excluding the newline character; by default it's
        initialized to ``min(terminal_width - 1, max_width)`` where
        ``max_width`` is another argument.
    :param max_width:
        maximum content line width (equivalent to ``Context.max_content_width``).
        Used to compute ``width`` when it is not provided, ignored otherwise.
    :param col1_max_width:
        the maximum width of the first column of a definition list; as in Click,
        if the text of a row exceeds this threshold, the 2nd column is printed
        on a new line.
    :param col2_min_width:
        the minimum width for the second column of a definition list; if the
        available space is less than this value, the formatter switches from the
        standard 2-column layout to the "linear layout" (that this decision
        is taken for each definition list). If you want to always use the linear
        layout, you can set this argument to a very high number (or ``math.inf``).
        If you never want it (not recommended), you can set this argument to zero.
    :param col_spacing:
        the number of spaces between the column boundaries of a definition list.
    :param row_sep:
        an "extra" separator to insert between the rows of a definition list (in
        addition to the normal newline between definitions). If you want an empty
        line between rows, pass ``row_sep=""``.
        Read :ref:`Row separators <row-separators>` for more.
    :param theme:
        an :class:`~cloup.HelpTheme` instance specifying how to style the various
        elements of the help page.
    """

    def __init__(
        self, indent_increment: int = 2,
        width: Optional[int] = None,
        max_width: Optional[int] = None,
        col1_max_width: int = 30,
        col2_min_width: int = 35,
        col_spacing: int = 2,
        row_sep: Union[None, str, 'SepGenerator', 'RowSepPolicy'] = None,
        theme: HelpTheme = HelpTheme(),
    ):
        check_positive_int(col1_max_width, 'col1_max_width')
        check_positive_int(col_spacing, 'col_spacing')
        if isinstance(row_sep, str) and row_sep.endswith('\n'):
            raise ValueError(
                "since v0.9, row_sep must not end with '\\n'. The formatter writes "
                "a '\\n' after it; no other newline is allowed.\n"
                "If you want an empty line between rows, set row_sep=''.")

        max_width = max_width or 80
        # We subtract 1 to the terminal width to leave space for the new line character.
        # Otherwise, when we write a line that is long exactly terminal_size (without \n)
        # the \n is printed on a new terminal line, leading to a useless empty line.
        width = (
            width
            or click.formatting.FORCED_WIDTH
            or min(max_width, shutil.get_terminal_size((80, 100)).columns - 1)
        )
        super().__init__(
            width=width, max_width=max_width, indent_increment=indent_increment
        )
        self.width: int = width
        self.col1_max_width = col1_max_width
        self.col2_min_width = col2_min_width
        self.col_spacing = col_spacing
        self.theme = theme
        self.row_sep = row_sep

    @staticmethod
    def settings(
        *, width: Possibly[Optional[int]] = MISSING,
        max_width: Possibly[Optional[int]] = MISSING,
        indent_increment: Possibly[int] = MISSING,
        col1_max_width: Possibly[int] = MISSING,
        col2_min_width: Possibly[int] = MISSING,
        col_spacing: Possibly[int] = MISSING,
        row_sep: Possibly[Union[None, str, 'SepGenerator', 'RowSepPolicy']] = MISSING,
        theme: Possibly[HelpTheme] = MISSING,
    ) -> Dict[str, Any]:
        """A utility method for creating a ``formatter_settings`` dictionary to
        pass as context settings or command attribute. This method exists for
        one only reason: it enables auto-complete for formatter options, thus
        improving the developer experience.

        Parameters are described in :class:`HelpFormatter`.
        """
        return pick_non_missing(locals())

    @property
    def available_width(self) -> int:
        return cast(int, self.width) - self.current_indent

    def write(self, *strings: str) -> None:
        self.buffer += strings

    def write_usage(
        self, prog: str, args: str = "", prefix: Optional[str] = 'Usage:'
    ) -> None:
        if prefix:
            prefix = self.theme.heading(prefix + ' ')
        prog = self.theme.invoked_command(prog)
        super().write_usage(prog, args, prefix)

    def write_aliases(self, aliases: Sequence[str]) -> None:
        self.write_heading("Aliases", newline=False)
        alias_list = ", ".join(self.theme.col1(alias) for alias in aliases)
        self.write(f" {alias_list}\n")

    def write_command_help_text(self, cmd: click.Command) -> None:
        help_text = cmd.help or ""
        if help_text and click_version_ge_8_1:
            help_text = inspect.cleandoc(help_text)
        if cmd.deprecated:
            help_text = "(DEPRECATED) " + help_text
        if help_text:
            self.write_paragraph()
            with self.indentation():
                self.write_text(help_text, style=self.theme.command_help)

    def write_heading(self, heading: str, newline: bool = True) -> None:
        if self.current_indent:
            self.write(" " * self.current_indent)
        self.write(self.theme.heading(heading + ":"))
        if newline:
            self.write('\n')

    def write_many_sections(
        self, sections: Sequence[HelpSection],
        aligned: bool = True,
    ) -> None:
        if aligned:
            return self.write_aligned_sections(sections)
        for s in sections:
            self.write_section(s)

    def write_aligned_sections(self, sections: Sequence[HelpSection]) -> None:
        """Writes multiple aligned definition lists."""
        all_rows = chain.from_iterable(dl.definitions for dl in sections)
        col1_width = self.compute_col1_width(all_rows, self.col1_max_width)
        for s in sections:
            self.write_section(s, col1_width=col1_width)

    def write_section(self, s: HelpSection, col1_width: Optional[int] = None) -> None:
        theme = self.theme
        self.write("\n")
        self.write_heading(s.heading, newline=not s.constraint)
        if s.constraint:
            constraint_text = f'[{s.constraint}]'
            available_width = self.available_width - len(s.heading) - len(': ')
            if len(constraint_text) <= available_width:
                self.write(" ", theme.constraint(constraint_text), "\n")
            else:
                self.write("\n")
                with self.indentation():
                    self.write_text(constraint_text, theme.constraint)

        with self.indentation():
            if s.help:
                self.write_text(s.help, theme.section_help)
            self.write_dl(s.definitions, col1_width=col1_width)

    def write_text(self, text: str, style: IStyle = identity) -> None:
        wrapped = wrap_text(
            text, self.width - self.current_indent, preserve_paragraphs=True)
        if style is identity:
            wrapped_text = textwrap.indent(wrapped, prefix=' ' * self.current_indent)
        else:
            styled_lines = map(style, wrapped.splitlines())
            lines = indent_lines(styled_lines, width=self.current_indent)
            wrapped_text = "\n".join(lines)
        self.write(wrapped_text, "\n")

    def compute_col1_width(self, rows: Iterable[Definition], max_width: int) -> int:
        col1_lengths = (unstyled_len(r[0]) for r in rows)
        lengths_under_limit = (length for length in col1_lengths if length <= max_width)
        return max(lengths_under_limit, default=0)

    def write_dl(  # type: ignore
        self, rows: Sequence[Definition],
        col_max: Optional[int] = None,  # default changed to None wrt parent class
        col_spacing: Optional[int] = None,  # default changed to None wrt parent class
        col1_width: Optional[int] = None,
    ) -> None:
        """Writes a definition list into the buffer. This is how options
        and commands are usually formatted.

        If there's enough space, definition lists are rendered as a 2-column
        pseudo-table: if the first column text of a row doesn't fit in the
        provided/computed ``col1_width``, the 2nd column is printed on the
        following line.

        If the available space for the 2nd column is below ``self.col2_min_width``,
        the 2nd "column" is always printed below the 1st, indented with a minimum
        of 3 spaces (or one ``indent_increment`` if that's greater than 3).

        :param rows:
            a list of two item tuples for the terms and values.
        :param col_max:
            the maximum width for the 1st column of a definition list; this
            argument is here to not break compatibility with Click; if provided,
            it overrides the attribute ``self.col1_max_width``.
        :param col_spacing:
            number of spaces between the first and second column;
            this argument is here to not break compatibility with Click;
            if provided, it overrides ``self.col_spacing``.
        :param col1_width:
            the width to use for the first column; if not provided, it's
            computed as the length of the longest string under ``self.col1_max_width``;
            useful when you need to align multiple definition lists.
        """
        # |<----------------------- width ------------------------>|
        # |                |<---------- available_width ---------->|
        # | current_indent | col1_width | col_spacing | col2_width |

        col1_max_width = min(
            col_max or self.col1_max_width,
            self.available_width,
        )
        col1_width = min(
            col1_width or self.compute_col1_width(rows, col1_max_width),
            col1_max_width,
        )
        col_spacing = col_spacing or self.col_spacing
        col2_width = self.available_width - col1_width - col_spacing

        if col2_width < self.col2_min_width:
            self.write_linear_dl(rows)
        else:
            self.write_tabular_dl(rows, col1_width, col_spacing, col2_width)

    def _get_row_sep_for(
        self, text_rows: Sequence[Sequence[str]],
        col_widths: Sequence[int],
        col_spacing: int,
    ) -> Optional[str]:
        if self.row_sep is None or isinstance(self.row_sep, str):
            return self.row_sep

        from .sep import RowSepPolicy
        if isinstance(self.row_sep, RowSepPolicy):
            return self.row_sep(text_rows, col_widths, col_spacing)
        elif callable(self.row_sep):  # RowSepPolicy is callable; keep this for last
            return self.row_sep(self.available_width)
        else:
            raise TypeError('row_sep')

    def write_tabular_dl(
        self, rows: Sequence[Definition],
        col1_width: int, col_spacing: int, col2_width: int,
    ) -> None:
        """Formats a definition list as a 2-column "pseudo-table". If the first
        column of a row exceeds ``col1_width``, the 2nd column is written on
        the subsequent line. This is the standard way of formatting definition
        lists and it's the default if there's enough space."""

        col1_plus_spacing = col1_width + col_spacing
        col2_indentation = " " * (
            self.current_indent + max(self.indent_increment, col1_plus_spacing)
        )
        indentation = " " * self.current_indent

        # Note: iter_defs() resolves eventual callables in row[1]
        text_rows = list(iter_defs(rows, col2_width))
        row_sep = self._get_row_sep_for(text_rows, (col1_width, col2_width), col_spacing)
        col1_styler, col2_styler = self.theme.col1, self.theme.col2

        def write_row(row):
            first, second = row
            self.write(indentation, col1_styler(first))
            if not second:
                self.write("\n")
            else:
                first_display_length = unstyled_len(first)
                if first_display_length <= col1_width:
                    spaces_to_col2 = col1_plus_spacing - first_display_length
                    self.write(" " * spaces_to_col2)
                else:
                    self.write("\n", col2_indentation)

                if len(second) <= col2_width:
                    self.write(col2_styler(second), "\n")
                else:
                    wrapped_text = wrap_text(second, col2_width, preserve_paragraphs=True)
                    lines = [col2_styler(line) for line in wrapped_text.splitlines()]
                    self.write(lines[0], "\n")
                    for line in lines[1:]:
                        self.write(col2_indentation, line, "\n")

        write_row(text_rows[0])
        for row in text_rows[1:]:
            if row_sep is not None:
                self.write(indentation, row_sep, "\n")
            write_row(row)

    def write_linear_dl(self, dl: Sequence[Definition]) -> None:
        """Formats a definition list as a "linear list". This is the default when
        the available width for the definitions (2nd column) is below
        ``self.col2_min_width``."""
        help_extra_indent = max(3, self.indent_increment)
        help_total_indent = self.current_indent + help_extra_indent
        help_max_width = self.width - help_total_indent
        current_indentation = " " * self.current_indent

        col1_styler = self.theme.col1
        col2_styler = self.theme.col2

        for names, help in iter_defs(dl, help_max_width):
            self.write(current_indentation + col1_styler(names) + '\n')
            if help:
                self.current_indent += help_extra_indent
                self.write_text(help, col2_styler)
                self.current_indent -= help_extra_indent
            self.write("\n")
        self.buffer.pop()  # pop last newline

    def write_epilog(self, epilog: str) -> None:
        self.write_text(epilog, self.theme.epilog)

    def __repr__(self):
        return make_repr(
            self, width=self.width, indent_increment=self.indent_increment,
            col1_max_width=self.col1_max_width, col_spacing=self.col_spacing
        )


def iter_defs(rows: Iterable[Definition], col2_width: int) -> Iterator[Tuple[str, str]]:
    for row in rows:
        if len(row) == 1:
            yield row[0], ''
        elif len(row) == 2:
            second = row[1](col2_width) if callable(row[1]) else row[1]
            yield row[0], second
        else:
            raise ValueError(f'invalid row length: {len(row)}')



"""
Constraints for parameter groups.

.. versionadded:: v0.5.0
"""
# flake8: noqa F401

from ._conditional import If
from ._core import (
    AcceptAtMost,
    AcceptBetween,
    And,
    Constraint,
    ErrorFmt,
    ErrorRephraser,
    HelpRephraser,
    Operator,
    Or,
    Rephraser,
    RequireAtLeast,
    RequireExactly,
    WrapperConstraint,
    accept_none,
    all_or_none,
    mutually_exclusive,
    require_all,
    require_any,
    require_one,
)
from ._support import (BoundConstraintSpec, ConstraintMixin, constrained_params, constraint)
from .conditions import AllSet, AnySet, Equal, IsSet, Not
from .exceptions import ConstraintViolated, UnsatisfiableConstraint



from collections import OrderedDict
from typing import (
    Dict, Iterable, List, Optional, Sequence, Tuple, Type, TypeVar, Union,
)

import click

from cloup._util import first_bool, pick_not_none
from cloup.formatting import HelpSection, ensure_is_cloup_formatter

CommandType = TypeVar('CommandType', bound=Type[click.Command])
Subcommands = Union[Iterable[click.Command], Dict[str, click.Command]]


class Section:
    """
    A group of (sub)commands to show in the same help section of a
    ``MultiCommand``. You can use sections with any `Command` that inherits
    from :class:`SectionMixin`.

    .. versionchanged:: 0.6.0
        removed the deprecated old name ``GroupSection``.

    .. versionchanged:: 0.5.0
        introduced the new name ``Section`` and deprecated the old ``GroupSection``.
    """

    def __init__(self, title: str,
                 commands: Subcommands = (),
                 sorted: bool = False):  # noqa
        """
        :param title:
        :param commands: sequence of commands or dictionary {name: command}
        :param sorted:
            if True, ``list_commands()`` returns the commands in lexicographic order
        """
        if not isinstance(title, str):
            raise TypeError(
                'the first argument must be a string, the title. You probably forgot it.')
        self.title = title
        self.sorted = sorted  # type: ignore
        self.commands: OrderedDict[str, click.Command] = OrderedDict()
        if isinstance(commands, Sequence):
            self.commands = OrderedDict()
            for cmd in commands:
                self.add_command(cmd)
        elif isinstance(commands, dict):
            self.commands = OrderedDict(commands)
        else:
            raise TypeError('the argument "commands" must be a list of commands '
                            'or a dict {name: command}')

    @classmethod
    def sorted(cls, title: str, commands: Subcommands = ()) -> 'Section':
        return cls(title, commands, sorted=True)

    def add_command(self, cmd: click.Command, name: Optional[str] = None):
        name = name or cmd.name
        if not name:
            raise TypeError('missing command name')
        if name in self.commands:
            raise Exception('command "{}" already exists'.format(name))
        self.commands[name] = cmd

    def list_commands(self) -> List[Tuple[str, click.Command]]:
        command_list = [(name, cmd) for name, cmd in self.commands.items()
                        if not cmd.hidden]
        if self.sorted:
            command_list.sort()
        return command_list

    def __len__(self) -> int:
        return len(self.commands)

    def __repr__(self) -> str:
        return 'Section({}, sorted={})'.format(self.title, self.sorted)


class SectionMixin:
    """
    Adds to a click.MultiCommand the possibility to organize its subcommands
    in multiple help sections.

    Sections can be specified in the following ways:

    #. passing a list of :class:`Section` objects to the constructor setting
       the argument ``sections``
    #. using :meth:`add_section` to add a single section
    #. using :meth:`add_command` with the argument `section` set

    Commands not assigned to any user-defined section are added to the
    "default section", whose title is "Commands" or "Other commands" depending
    on whether it is the only section or not. The default section is the last
    shown section in the help and its commands are listed in lexicographic order.

    .. versionchanged:: 0.8.0
        this mixin now relies on ``cloup.HelpFormatter`` to align help sections.
        If a ``click.HelpFormatter`` is used with a ``TypeError`` is raised.

    .. versionchanged:: 0.8.0
        removed ``format_section``. Added ``make_commands_help_section``.

    .. versionadded:: 0.5.0
    """

    def __init__(
        self, *args,
        commands: Optional[Dict[str, click.Command]] = None,
        sections: Iterable[Section] = (),
        align_sections: Optional[bool] = None,
        **kwargs,
    ):
        """
        :param align_sections:
            whether to align the columns of all subcommands' help sections.
            This is also available as a context setting having a lower priority
            than this attribute. Given that this setting should be consistent
            across all you commands, you should probably use the context
            setting only.
        :param args:
            positional arguments forwarded to the next class in the MRO
        :param kwargs:
            keyword arguments forwarded to the next class in the MRO
        """
        self.align_sections = align_sections
        self._default_section = Section('__DEFAULT', commands=commands or [])
        self._user_sections: List[Section] = []
        self._section_set = {self._default_section}
        for section in sections:
            self.add_section(section)
        super().__init__(*args, commands=commands, **kwargs)  # type: ignore

    def _add_command_to_section(self, cmd, name=None, section=None):
        """Adds a command to the section (if specified) or to the default section."""
        name = name or cmd.name
        if section is None:
            section = self._default_section
        section.add_command(cmd, name)
        if section not in self._section_set:
            self._user_sections.append(section)
            self._section_set.add(section)

    def add_section(self, section: Section) -> None:
        """ Adds a :class:`Section` to this group. You can add the same
        section object a single time. """
        if section in self._section_set:
            raise ValueError(f'{section} was already added')
        self._user_sections.append(section)
        self._section_set.add(section)
        for name, cmd in section.commands.items():
            # It's important to call self.add_command() and not super().add_command() here
            # otherwise subclasses' add_command() is not called.
            self.add_command(cmd, name, fallback_to_default_section=False)

    def section(self, title: str, *commands: click.Command, **attrs) -> Section:
        """ Creates a new :class:`Section`, adds it to this group and returns it."""
        section = Section(title, commands, **attrs)
        self.add_section(section)
        return section

    def add_command(
        self, cmd: click.Command,
        name: Optional[str] = None,
        section: Optional[Section] = None,
        fallback_to_default_section: bool = True,
    ) -> None:
        """
        Adds a subcommand to this ``Group``.

        **Implementation note:** ``fallback_to_default_section`` looks not very
        clean but, even if it's not immediate to see (it wasn't for me), I chose
        it over apparently cleaner options.

        :param cmd:
        :param name:
        :param section:
            a ``Section`` instance. The command must not be in the section already.
        :param fallback_to_default_section:
            if ``section`` is None and this option is enabled, the command is added
            to the "default section". If disabled, the command is not added to
            any section unless ``section`` is provided. This is useful for
            internal code and subclasses. Don't disable it unless you know what
            you are doing.
        """
        super().add_command(cmd, name)  # type: ignore
        if section or fallback_to_default_section:
            self._add_command_to_section(cmd, name, section)

    def list_sections(self, ctx: click.Context,
                      include_default_section: bool = True) -> List[Section]:
        """ Returns the list of all sections in the "correct order".
         if ``include_default_section=True`` and the default section is non-empty,
         it will be included at the end of the list. """
        section_list = list(self._user_sections)
        if include_default_section and len(self._default_section) > 0:
            default_section = Section.sorted(
                title='Other commands' if len(self._user_sections) > 0 else 'Commands',
                commands=self._default_section.commands)
            section_list.append(default_section)
        return section_list

    def format_subcommand_name(
        self, ctx: click.Context, name: str, cmd: click.Command
    ) -> str:
        """Used to format the name of the subcommands. This method is useful
        when you combine this extension with other click extensions that override
        :meth:`format_commands`. Most of these, like click-default-group, just
        add something to the name of the subcommands, which is exactly what this
        method allows you to do without overriding bigger methods.
        """
        return name

    def make_commands_help_section(
        self, ctx: click.Context, section: Section
    ) -> Optional[HelpSection]:
        visible_subcommands = section.list_commands()
        if not visible_subcommands:
            return None
        return HelpSection(
            heading=section.title,
            definitions=[
                (self.format_subcommand_name(ctx, name, cmd), cmd.get_short_help_str)
                for name, cmd in visible_subcommands
            ]
        )

    def must_align_sections(
        self, ctx: Optional[click.Context], default: bool = True
    ) -> bool:
        return first_bool(
            self.align_sections,
            getattr(ctx, 'align_sections', None),
            default,
        )

    def format_commands(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        formatter = ensure_is_cloup_formatter(formatter)

        subcommand_sections = self.list_sections(ctx)
        help_sections = pick_not_none(
            self.make_commands_help_section(ctx, section)
            for section in subcommand_sections
        )
        if not help_sections:
            return

        formatter.write_many_sections(
            help_sections, aligned=self.must_align_sections(ctx)
        )



from unittest.mock import Mock

import click
import pytest
from click import Argument, Option

import cloup
from cloup import Context
from cloup._util import pick_non_missing, reindent
from cloup.constraints import Constraint, RequireAtLeast, mutually_exclusive
from cloup.typing import MISSING
from tests.constraints.test_constraints import FakeConstraint
from tests.util import new_dummy_func, pick_first_bool


class TestConstraintMixin:
    def test_params_are_correctly_grouped_by_name(self):
        params = [
            Argument(('arg1',)),
            Option(('--str-opt',)),
            Option(('--int-opt', 'option2')),
        ]
        cmd = cloup.Command(name='cmd', params=params, callback=new_dummy_func())
        for param in params:
            assert cmd.get_param_by_name(param.name) == param

        with pytest.raises(KeyError):
            cmd.get_param_by_name('non-existing')

        assert cmd.get_params_by_name(['arg1', 'option2']) == (params[0], params[2])


@pytest.mark.parametrize('command_type', ["command", "group"])
@pytest.mark.parametrize('do_check_consistency', [
    pytest.param(True, id="with_consistency_checks"),
    pytest.param(False, id="without_consistency_checks")
])
def test_constraints_are_checked_according_to_protocol(
    runner, command_type, do_check_consistency
):
    constraints = [
        Mock(spec_set=Constraint, wraps=FakeConstraint())
        for _ in range(3)
    ]
    settings = Context.settings(check_constraints_consistency=do_check_consistency)

    command_decorator = cloup.group if command_type == "group" else cloup.command

    @command_decorator(context_settings=settings)
    @cloup.option_group('first', cloup.option('--a'), cloup.option('--b'),
                        constraint=constraints[0])
    @cloup.option_group('second', cloup.option('--c'), cloup.option('--d'),
                        constraint=constraints[1])
    @cloup.constraint(constraints[2], ['a', 'c'])
    @cloup.pass_context
    def cmd(ctx, a, b, c, d):
        assert Constraint.must_check_consistency(ctx) == do_check_consistency
        print(f'{a}, {b}, {c}, {d}')

    shell = '--a=1 --c=2'
    if isinstance(cmd, cloup.Group):
        cmd.add_command(cloup.Command(name="dummy", callback=lambda: 0))
        shell += ' dummy'

    result = runner.invoke(cmd, args=shell.split())

    assert result.output.strip() == '1, None, 2, None'
    for constr, opt_names in zip(constraints, [['a', 'b'], ['c', 'd'], ['a', 'c']]):
        opts = cmd.get_params_by_name(opt_names)
        if do_check_consistency:
            constr.check_consistency.assert_called_once_with(opts)
        else:
            constr.check_consistency.assert_not_called()
        constr.check_values.assert_called_once()


@pytest.mark.parametrize('command_type', ["command", "group"])
@pytest.mark.parametrize(
    'cmd_value', [MISSING, None, True, False],
    ids=lambda val: f'cmd_{val}'
)
@pytest.mark.parametrize(
    'ctx_value', [MISSING, None, True, False],
    ids=lambda val: f'ctx_{val}'
)
def test_constraints_are_shown_in_help_only_if_feature_is_enabled(
    runner, command_type, cmd_value, ctx_value
):
    should_show = pick_first_bool([cmd_value, ctx_value], default=False)
    cxt_settings = pick_non_missing(dict(
        show_constraints=ctx_value,
        terminal_width=80,
    ))
    cmd_kwargs = pick_non_missing(dict(
        show_constraints=cmd_value,
        context_settings=cxt_settings
    ))

    command_decorator = cloup.group if command_type == "group" else cloup.command

    @command_decorator(**cmd_kwargs)
    @cloup.option('--a')
    @cloup.option('--b')
    @cloup.option('--c')
    @cloup.constraint(FakeConstraint(help='a constraint'), ['a', 'b'])
    @cloup.constraint(FakeConstraint(help='another constraint'), ['b', 'c'])
    def cmd(a, b, c, d):
        pass

    if isinstance(cmd, cloup.Group):
        cmd.add_command(cloup.Command(name="dummy", callback=lambda: 0))

    result = runner.invoke(cmd, args=['--help'],
                           catch_exceptions=False,
                           prog_name='test')
    out = result.output

    if command_type == "group":
        if should_show:
            assert out == reindent("""
                Usage: test [OPTIONS] COMMAND [ARGS]...

                Options:
                  --a TEXT
                  --b TEXT
                  --c TEXT
                  --help    Show this message and exit.

                Constraints:
                  {--a, --b}  a constraint
                  {--b, --c}  another constraint

                Commands:
                  dummy
            """)
        else:
            assert out == reindent("""
                Usage: test [OPTIONS] COMMAND [ARGS]...

                Options:
                  --a TEXT
                  --b TEXT
                  --c TEXT
                  --help    Show this message and exit.

                Commands:
                  dummy
            """)
    else:
        base_help = reindent("""
            Usage: test [OPTIONS]

            Options:
              --a TEXT
              --b TEXT
              --c TEXT
              --help    Show this message and exit.
        """)

        if should_show:
            constraints_section = reindent("""
                Constraints:
                  {--a, --b}  a constraint
                  {--b, --c}  another constraint
            """)
            expected = base_help + "\n" + constraints_section
            assert out == expected
        else:
            assert out == base_help


def test_usage_of_constraints_as_decorators(runner):
    require_any = RequireAtLeast(1)

    @cloup.command()
    @require_any(
        cloup.argument('arg', required=False),
        cloup.option('-a'),
        cloup.option('-b'),
    )
    @mutually_exclusive(
        cloup.option('-c'),
        cloup.option('-d'),
    )
    def cmd(arg, a, b, c, d):
        pass

    assert runner.invoke(cmd, args='ARG -c CCC'.split()).exit_code == 0
    assert runner.invoke(cmd, args='-a AAA -d DDD'.split()).exit_code == 0

    res = runner.invoke(cmd, args=[])
    assert res.exit_code == click.UsageError.exit_code
    assert 'at least 1 of the following' in res.output

    res = runner.invoke(cmd, args='ARG -c CCC -d DDD'.split())
    assert res.exit_code == click.UsageError.exit_code
    assert 'mutually exclusive' in res.output



import random
from pathlib import Path

random.seed(12345)

EXAMPLE_FILE_PATH = Path(__file__).parent.parent / 'examples/git_sections.py'
MAX_SECTION_COUNT = 3
MAX_COMMANDS_PER_SECTION = 3

GIT_HELP = """
start a working area (see also: git help tutorial)
   clone             Clone a repository into a new directory
   init              Create an empty Git repository or reinitialize an existing one

work on the current change (see also: git help everyday)
   add               Add file contents to the index
   mv                Move or rename a file, a directory, or a symlink
   restore           Restore working tree files
   rm                Remove files from the working tree and from the index
   sparse-checkout   Initialize and modify the sparse-checkout

examine the history and state (see also: git help revisions)
   bisect            Use binary search to find the commit that introduced a bug
   diff              Show changes between commits, commit and working tree, etc
   grep              Print lines matching a pattern
   log               Show commit logs
   show              Show various types of objects
   status            Show the working tree status

grow, mark and tweak your common history
   branch            List, create, or delete branches
   commit            Record changes to the repository
   merge             Join two or more development histories together
   rebase            Reapply commits on top of another base tip
   reset             Reset current HEAD to the specified state
   switch            Switch branches
   tag               Create, list, delete or verify a tag object signed with GPG
""".strip()

CODE_TEMPLATE = """
\"\"\"
This example shows how to use cloup.Section to organize the subcommands of
a multi-command in many --help sections.

The code was generated by parsing "git --help" and taking
- the first {max_section_count} sections
- for each section, the first {max_commands_per_section} commands after shuffling them.
\"\"\"
import cloup


def f(**kwargs):
    \"\"\" Dummy command callback \"\"\"
    print(**kwargs)


# In a real big application, you would import the following commands from separate modules =========
{command_list}
# ==================================================================================================

\"\"\"
If "align_sections=True" (default), the help column of all sections will
be aligned; otherwise, each section will be formatted independently.
\"\"\"
@cloup.group('git', align_sections=True)
def git():
    return 0

\"\"\"
git.section() creates a new Section object, adds it to git and returns it.

In the help, sections are shown in the same order they are added.
Commands in each sections are shown in the same order they are listed, unless
you pass the argument "sorted_=True".
\"\"\"
{section_list}

\"\"\"
In alternative, you can either:
- pass a list of Section objects as argument "sections" to cloup.Group or @cloup.group
- use git.add_section(section) to add an existing Section object
- use git.add_command(cmd, name, section, ...); the section must NOT contain the command
- use @git.command(cmd, name, section, ...)

Individual commands don't store the section they belong to.
Also, cloup.command doesn't accept a "section" argument.
\"\"\"

# The following commands will be added to the "default section" (a sorted Section)
git.add_command(cloup.command('fake-2', help='Fake command #2')(f))
git.add_command(cloup.command('fake-1', help='Fake command #1')(f))


if __name__ == '__main__':
    git()
""".lstrip()


def parse_help(help_text):
    sections = []
    for text in help_text.split('\n\n'):
        title, *rows = map(str.strip, text.split('\n'))
        title = title.strip().capitalize()
        commands = []
        for row in rows:
            cmd, desc = map(str.strip, row.split(' ', 1))
            commands.append((cmd, desc))
        random.shuffle(commands)
        sections.append([title, commands])
    return sections


def get_command_var_name(cmd_name):
    return 'git_' + cmd_name.replace('-', '_')


def generate(out_path=EXAMPLE_FILE_PATH,
             max_section_count=MAX_SECTION_COUNT,
             max_commands_per_section=MAX_COMMANDS_PER_SECTION):
    sections = parse_help(GIT_HELP)[:max_section_count]
    for s in sections:
        s[1] = s[1][:max_commands_per_section]

    # Subcommand definitions
    commands_buffer = []
    for _, commands in sections:
        for cmd_name, desc in commands:
            var_name = get_command_var_name(cmd_name)
            commands_buffer.append(
                f'{var_name} = cloup.command({cmd_name!r}, help={desc!r})(f)')
        commands_buffer.append('')
    commands_list = '\n'.join(commands_buffer).strip()

    # Section list
    sections_buffer = []
    for title, commands in sections:
        sections_buffer.append("git.section(")
        sections_buffer.append(f"    {title!r},")
        for cmd_name, _ in commands:
            var_name = get_command_var_name(cmd_name)
            sections_buffer.append(f"    {var_name},")
        sections_buffer.append(')')
    section_list = '\n'.join(sections_buffer)

    code = CODE_TEMPLATE.format(max_section_count=max_section_count,
                                max_commands_per_section=max_commands_per_section,
                                command_list=commands_list,
                                section_list=section_list)

    print(code)
    with open(out_path, 'w', encoding='utf-8') as out_file:
        out_file.write(code)


if __name__ == '__main__':
    generate()



import shutil

import click
from click import argument, command


@command(no_args_is_help=True)
@argument('src', type=click.Path(exists=True))
@argument('dest', type=click.Path(file_okay=False))
def f(src: str, dest: str):
    """Copy a SRC file/folder to DEST with merging."""
    shutil.copytree(src, dest, dirs_exist_ok=True)


if __name__ == '__main__':
    f()



import inspect

import click
import pytest

from cloup.styling import Color, HelpTheme, Style


def test_help_theme_copywith():
    s1, s2 = Style(), Style()
    r1, r2 = Style(), Style()
    theme = HelpTheme(heading=s1, col1=s2).with_(col1=r1, col2=r2)
    assert theme == HelpTheme(heading=s1, col1=r1, col2=r2)


def test_help_theme_copywith_takes_the_same_parameters_of_constructor():
    def get_param_names(func):
        params = inspect.signature(func).parameters
        return list(params.keys())

    constructor_params = get_param_names(HelpTheme)
    method_params = get_param_names(HelpTheme.with_)[1:]  # skip self
    assert method_params == constructor_params


def test_help_theme_default_themes():
    assert isinstance(HelpTheme.dark(), HelpTheme)
    assert isinstance(HelpTheme.light(), HelpTheme)


def test_style():
    text = 'hi there'
    kwargs = dict(fg=Color.green, bold=True, blink=True)
    assert Style(**kwargs)(text) == click.style(text, **kwargs)


def test_unsupported_style_args_are_ignored_in_click_7():
    Style(overline=True, italic=True, strikethrough=True)


def test_Color():
    assert Color.red == 'red'
    assert Color.bright_blue == 'bright_blue'

    with pytest.raises(Exception, match="it's not instantiable"):
        Color()
    with pytest.raises(Exception, match="you can't set attributes on this class"):
        Color.red = 'blue'

    assert 'red' in Color
    assert 'pippo' not in Color
    d = Color.asdict()
    for k, v in d.items():
        assert k in Color
        assert Color[k] == v



"""Generic utilities."""
from typing import (
    Any, Dict, Hashable, Iterable, List, Optional, Sequence, TypeVar,
)

import click

from cloup.typing import MISSING, Possibly

click_version_tuple = tuple(click.__version__.split('.'))
click_major = int(click_version_tuple[0])
click_minor = int(click_version_tuple[1])
click_version_ge_8_1 = (click_major, click_minor) >= (8, 1)

T = TypeVar('T')
K = TypeVar('K', bound=Hashable)
V = TypeVar('V')


def pick_non_missing(d: Dict[K, Possibly[V]]) -> Dict[K, V]:
    return {key: val for key, val in d.items() if val is not MISSING}


def class_name(obj):
    return obj.__class__.__name__


def check_arg(condition: bool, msg: str = ''):
    if not condition:
        raise ValueError(msg)


def indent_lines(lines: Iterable[str], width=2) -> List[str]:
    spaces = ' ' * width
    return [spaces + line for line in lines]


def make_repr(obj, *args, _line_len: int = 60, _indent: int = 2, **kwargs) -> str:
    """
    Generate repr(obj).

    :param obj:
        object to represent
    :param args:
        positional arguments in the repr
    :param _line_len:
        if the repr length exceeds this, arguments will be on their own line;
        if negative, the repr will be in a single line regardless of its length
    :param _indent:
        indentation width of arguments in case they are shown in their own line
    :param kwargs:
        keyword arguments in the repr
    :return: str
    """
    cls_name = obj.__class__.__name__
    arglist = [
        *(repr(arg) for arg in args),
        *(f'{key}={value!r}' for key, value in kwargs.items()),
    ]
    len_arglist = sum(len(s) for s in arglist)
    total_len = len(cls_name) + len_arglist + 2 * len(arglist)
    if 0 <= _line_len < total_len:
        lines = indent_lines(arglist, width=_indent)
        args_text = ',\n'.join(lines)
        return f'{cls_name}(\n{args_text}\n)'
    else:
        args_text = ', '.join(arglist)
        return f'{cls_name}({args_text})'


def make_one_line_repr(obj, *args, **kwargs):
    return make_repr(obj, *args, _line_len=-1, **kwargs)


def pluralize(
    count: int, zero: str = '', one: str = '', many: str = '',
) -> str:
    if count == 0 and zero:
        return zero
    if count == 1 and one:
        return one
    return many.format(count=count)


def coalesce(*values: Optional[T]) -> Optional[T]:
    """Returns the first value that is not None (or None if no such value exists)."""
    return next((val for val in values if val is not None), None)


def first_bool(*values: Any) -> bool:
    """Returns the first bool (or raises StopIteration if no bool is found)."""
    return next(val for val in values if isinstance(val, bool))


def pick_not_none(iterable: Iterable[Optional[T]]) -> List[T]:
    return [x for x in iterable if x is not None]


def check_positive_int(value, arg_name):
    error_type = None
    if not isinstance(value, int):
        error_type = TypeError
    elif value <= 0:
        error_type = ValueError
    if error_type:
        raise error_type(f'{arg_name} should be a positive integer. It is: {value}.')


def identity(x: T) -> T:
    return x


class FrozenSpaceMeta(type):
    def __init__(cls, *args):
        d = {k: v for k, v in vars(cls).items() if not k.startswith('_')}
        type.__setattr__(cls, '_dict', d)

    def __setattr__(cls, key, value):
        raise Exception("you can't set attributes on this class")

    def asdict(cls) -> Dict[str, Any]:
        return cls._dict  # type: ignore

    def __contains__(cls, item: str) -> bool:
        return item in cls.asdict()

    def __getitem__(cls, item):
        return cls._dict[item]


class FrozenSpace(metaclass=FrozenSpaceMeta):
    """A class used just as frozen namespace for constants."""

    def __init__(self):
        raise Exception(
            "this class is just a namespace for constants, it's not instantiable.")


def delete_keys(d: dict, keys: Sequence[str]) -> None:
    for key in keys:
        del d[key]


def reindent(text: str, indent: int = 0) -> str:
    import textwrap as tw
    if text.startswith('\n'):
        text = text[1:]
    text = tw.dedent(text)
    if indent:
        return tw.indent(text, ' ' * indent)
    return text



# flake8: noqa E128
from typing import cast

import click

import cloup
from cloup import Command, HelpFormatter, argument, option, option_group
from cloup.constraints import AcceptAtMost, If, RequireAtLeast


def make_example_command(
    align_option_groups: bool,
    tabular_help: bool = True,
) -> Command:
    @cloup.command(
        'clouptest',
        align_option_groups=align_option_groups,
        formatter_settings=HelpFormatter.settings(
            width=80,
            col2_min_width=30 if tabular_help else 80,
        ),
        epilog='Made with love by Gianluca.'
    )
    @argument("arg_one", help="This is the description of argument #1.")
    @argument("arg_two", help="This is the description of argument #2.", required=False)
    @option_group(
        'Option group A',
        option('--one', help='The one thing you need to run this command.'),
        option('--two', help='This is long description that should be wrapped into '
                             'multiple lines so that the entire text stays inside '
                             'the allowed width.'),
        option('--three', help='The 3rd option of group A.'),
        help="This is a very useful description of group A. This is a rarely used "
             "feature but, as the others, needs to be tested. I'm making this "
             "unnecessarily long in order to test wrapping.",
        constraint=AcceptAtMost(2)
    )
    @option_group(
        'Option group B',
        'Help as positional argument.',
        option('--four / --no-four', help='The 1st option of group B.'),
        option('--five', help='The 2nd option of group B.', hidden=True),  # hidden option
        option('--six', help='The 3rd option of group B.'),
        constraint=If('three', then=RequireAtLeast(1))
    )
    @option('--seven', help='First uncategorized option.',
            type=click.Choice('yes no ask'.split()))
    @option('--height', help='Second uncategorized option.')
    @option('--nine', help='Third uncategorized option.', hidden=True)
    def cmd(**kwargs):
        """ A CLI that does nothing. """
        print(kwargs)

    if tabular_help:
        expected_help = (_TABULAR_ALIGNED_HELP if align_option_groups
                         else _TABULAR_NON_ALIGNED_HELP)
    else:
        expected_help = _LINEAR_HELP

    cmd.expected_help = expected_help   # type: ignore
    return cast(Command, cmd)


_TABULAR_ALIGNED_HELP = """
Usage: clouptest [OPTIONS] ARG_ONE [ARG_TWO]

  A CLI that does nothing.

Positional arguments:
  ARG_ONE               This is the description of argument #1.
  [ARG_TWO]             This is the description of argument #2.

Option group A: [at most 2 accepted]
  This is a very useful description of group A. This is a rarely used feature
  but, as the others, needs to be tested. I'm making this unnecessarily long in
  order to test wrapping.
  --one TEXT            The one thing you need to run this command.
  --two TEXT            This is long description that should be wrapped into
                        multiple lines so that the entire text stays inside the
                        allowed width.
  --three TEXT          The 3rd option of group A.

Option group B: [at least 1 required if --three is set]
  Help as positional argument.
  --four / --no-four    The 1st option of group B.
  --six TEXT            The 3rd option of group B.

Other options:
  --seven [yes|no|ask]  First uncategorized option.
  --height TEXT         Second uncategorized option.
  --help                Show this message and exit.

Made with love by Gianluca.
""".strip()

_TABULAR_NON_ALIGNED_HELP = """
Usage: clouptest [OPTIONS] ARG_ONE [ARG_TWO]

  A CLI that does nothing.

Positional arguments:
  ARG_ONE    This is the description of argument #1.
  [ARG_TWO]  This is the description of argument #2.

Option group A: [at most 2 accepted]
  This is a very useful description of group A. This is a rarely used feature
  but, as the others, needs to be tested. I'm making this unnecessarily long in
  order to test wrapping.
  --one TEXT    The one thing you need to run this command.
  --two TEXT    This is long description that should be wrapped into multiple
                lines so that the entire text stays inside the allowed width.
  --three TEXT  The 3rd option of group A.

Option group B: [at least 1 required if --three is set]
  Help as positional argument.
  --four / --no-four  The 1st option of group B.
  --six TEXT          The 3rd option of group B.

Other options:
  --seven [yes|no|ask]  First uncategorized option.
  --height TEXT         Second uncategorized option.
  --help                Show this message and exit.

Made with love by Gianluca.
""".strip()

_LINEAR_HELP = """
Usage: clouptest [OPTIONS] ARG_ONE [ARG_TWO]

  A CLI that does nothing.

Positional arguments:
  ARG_ONE
     This is the description of argument #1.

  [ARG_TWO]
     This is the description of argument #2.

Option group A: [at most 2 accepted]
  This is a very useful description of group A. This is a rarely used feature
  but, as the others, needs to be tested. I'm making this unnecessarily long in
  order to test wrapping.
  --one TEXT
     The one thing you need to run this command.

  --two TEXT
     This is long description that should be wrapped into multiple lines so that
     the entire text stays inside the allowed width.

  --three TEXT
     The 3rd option of group A.

Option group B: [at least 1 required if --three is set]
  Help as positional argument.
  --four / --no-four
     The 1st option of group B.

  --six TEXT
     The 3rd option of group B.

Other options:
  --seven [yes|no|ask]
     First uncategorized option.

  --height TEXT
     Second uncategorized option.

  --help
     Show this message and exit.

Made with love by Gianluca.
""".strip()


if __name__ == '__main__':
    make_example_command(align_option_groups=False, tabular_help=True)(
        ['--help'], prog_name='clouptest'
    )



import pathlib

import click

import cloup


def test_path():
    p = cloup.path()
    assert isinstance(p, click.Path)
    assert p.type == pathlib.Path


def test_dir_path():
    p = cloup.dir_path()
    assert isinstance(p, click.Path)
    assert p.type == pathlib.Path
    assert not p.file_okay
    assert p.dir_okay


def test_file_path():
    p = cloup.file_path()
    assert isinstance(p, click.Path)
    assert p.type == pathlib.Path
    assert not p.dir_okay
    assert p.file_okay



import click
import pytest

import cloup
from cloup import Group
from cloup._util import first_bool, reindent
from cloup.typing import MISSING


@pytest.fixture()
def cli() -> cloup.Group:
    @cloup.group()
    def cli():
        """A package installer."""
        pass

    @cloup.command(aliases=['i', 'add'])
    @cloup.argument('pkg')
    def install(pkg: str):
        """Install a package."""
        print('install', pkg)

    @cloup.group(aliases=['conf', 'cfg'])
    def config():
        """Manage the configuration."""
        print('config')

    @cloup.command(aliases=['clr'], cls=click.Command)
    def clear():
        """Remove all installed packages."""
        print('clear')

    cli.section('Commands', install, clear, config)
    return cli


cli_help_without_aliases = reindent("""
    Usage: cli [OPTIONS] COMMAND [ARGS]...

      A package installer.

    Options:
      --help  Show this message and exit.

    Commands:
      install  Install a package.
      clear    Remove all installed packages.
      config   Manage the configuration.
""")

cli_help_with_aliases = reindent("""
    Usage: cli [OPTIONS] COMMAND [ARGS]...

      A package installer.

    Options:
      --help  Show this message and exit.

    Commands:
      install (i, add)    Install a package.
      clear (clr)         Remove all installed packages.
      config (conf, cfg)  Manage the configuration.
""")


def test_command_aliases_are_stored_in_the_command(cli):
    assert cli.commands['install'].aliases == ['i', 'add']
    assert cli.commands['clear'].aliases == ['clr']


def test_simple_command_name_resolution(cli):
    ctx = cloup.Context(command=cli)
    assert cli.resolve_command_name(ctx, 'install') == 'install'
    assert cli.resolve_command_name(ctx, 'i') == 'install'
    assert cli.resolve_command_name(ctx, 'add') == 'install'

    assert cli.resolve_command_name(ctx, 'config') == 'config'
    assert cli.resolve_command_name(ctx, 'conf') == 'config'

    assert cli.resolve_command_name(ctx, 'clear') == 'clear'
    assert cli.resolve_command_name(ctx, 'clr') == 'clear'


def test_command_name_resolution_with_token_normalization_function(cli):
    ctx = cloup.Context(command=cli, token_normalize_func=str.lower)
    assert cli.resolve_command_name(ctx, 'INSTALL') == 'install'
    assert cli.resolve_command_name(ctx, 'ADD') == 'install'
    assert cli.resolve_command_name(ctx, 'CLR') == 'clear'
    assert cli.resolve_command_name(ctx, 'cONF') == 'config'


@pytest.mark.parametrize('alias', ['install', 'i', 'add'])
def test_command_resolution_with_cloup_subcommand(cli, runner, alias):
    res = runner.invoke(cli, [alias, 'cloup'])
    assert res.output.strip() == 'install cloup'


@pytest.mark.parametrize('alias', ['clear', 'clr'])
def test_command_resolution_with_click_subcommand(cli, runner, alias):
    res = runner.invoke(cli, [alias])
    assert res.output.strip() == 'clear'


@pytest.mark.parametrize(
    'cmd_value', [MISSING, None, True, False],
    ids=lambda val: f'cmd_{val}'
)
@pytest.mark.parametrize(
    'ctx_value', [MISSING, None, True, False],
    ids=lambda val: f'ctx_{val}'
)
def test_show_subcommand_aliases_setting(cli, runner, ctx_value, cmd_value):
    if ctx_value is not MISSING:
        cli.context_settings['show_subcommand_aliases'] = ctx_value
    if cmd_value is not MISSING:
        cli.show_subcommand_aliases = cmd_value

    should_show_aliases = first_bool(cmd_value, ctx_value, Group.SHOW_SUBCOMMAND_ALIASES)
    expected_help = (cli_help_with_aliases
                     if should_show_aliases
                     else cli_help_without_aliases)

    res = runner.invoke(cli, ['--help'])
    assert res.output == expected_help


def test_cloup_subcommand_help(cli, runner):
    res = runner.invoke(cli, ['i', '--help'])
    # 1. Shows the full subcommand name even if an alias was used.
    # 2. Shows aliases after help text.
    expected = reindent("""
        Usage: cli install [OPTIONS] PKG
        Aliases: i, add

          Install a package.

        Options:
          --help  Show this message and exit.
    """)
    assert res.output == expected


def test_click_subcommand_help(cli, runner):
    res = runner.invoke(cli, ['clr', '--help'])
    # Shows the full subcommand name even if an alias was used.
    # Aliases are not shown (need Cloup commands for that).
    expected = reindent("""
        Usage: cli clear [OPTIONS]

          Remove all installed packages.

        Options:
          --help  Show this message and exit.
    """)
    assert res.output == expected


def test_cloup_subgroup_help(cli, runner):
    res = runner.invoke(cli, ['conf', '--help'])
    # 1. Shows the full subcommand name even if an alias was used.
    # 2. Shows aliases after help text.
    expected = reindent("""
        Usage: cli config [OPTIONS] COMMAND [ARGS]...
        Aliases: conf, cfg

          Manage the configuration.

        Options:
          --help  Show this message and exit.
    """)
    assert res.output == expected



from functools import partial
from typing import Sequence
from unittest import mock
from unittest.mock import Mock

import click
import pytest
from click import Command, Context, Parameter
from pytest import mark

from cloup.constraints import (
    AcceptAtMost,
    AcceptBetween,
    Constraint,
    ErrorFmt,
    Rephraser,
    RequireAtLeast,
    RequireExactly,
    require_all,
)
from cloup.constraints.exceptions import ConstraintViolated, UnsatisfiableConstraint
from tests.util import (
    make_context, make_fake_context, make_options, parametrize, should_raise
)


class FakeConstraint(Constraint):
    """Sometimes it's useful to use::

        Mock(wraps=FakeConstraint(...))

    to create a test double with characteristics of both a mock and a fake."""

    def __init__(self, satisfied=True, consistent=True, help='help',
                 error='error', inconsistency_reason='consistency_error'):
        self.satisfied = satisfied
        self.consistent = consistent
        self._help = help
        self.error = error
        self.inconsistency_reason = inconsistency_reason
        self.check_consistency_calls = []
        self.check_values_calls = []

    def help(self, ctx: Context) -> str:
        return self._help

    def check_consistency(self, params: Sequence[Parameter]) -> None:
        self.check_consistency_calls.append(dict(params=params))
        if not self.consistent:
            raise UnsatisfiableConstraint(self, params, self.inconsistency_reason)

    def check_values(self, params: Sequence[Parameter], ctx: Context):
        self.check_values_calls.append(dict(params=params, ctx=ctx))
        if not self.satisfied:
            raise ConstraintViolated(self.error, ctx=ctx, constraint=self, params=params)


class TestBaseConstraint:
    def test_rephrased_calls_Rephraser_correctly(self):
        with mock.patch('cloup.constraints._core.Rephraser') as rephraser_cls:
            cons = FakeConstraint()
            cons.rephrased(help='ciao')
            rephraser_cls.assert_called_with(cons, help='ciao', error=None)
            cons.rephrased(error='ciao')
            rephraser_cls.assert_called_with(cons, help=None, error='ciao')

    def test_hidden_constraint_returns_empty_help(self, dummy_ctx):
        hidden = FakeConstraint(help='non-empty help').hidden()
        assert isinstance(hidden, Rephraser)
        assert hidden.help(dummy_ctx) == ''

    @mark.parametrize('satisfied', [True, False])
    @mark.parametrize('consistent', [True, False])
    def test__call__raises_iff_check_raises(self, satisfied, consistent):
        ctx = make_fake_context(make_options('abc'))
        cons = FakeConstraint(satisfied=satisfied, consistent=consistent)
        exc_class = UnsatisfiableConstraint if not consistent else ConstraintViolated
        with should_raise(exc_class, when=not (consistent and satisfied)):
            cons.check(['a', 'b'], ctx)

    @parametrize(
        ['ctx_kwargs', 'should_check'],
        pytest.param(dict(cls=click.Context), True, id='click.Context [no setting]'),
        pytest.param(dict(), True, id='cloup.Context [default]'),
        pytest.param(dict(check_constraints_consistency=False), False, id='disabled'),
    )
    def test_check_consistency_is_called_unless_disabled(self, ctx_kwargs, should_check):
        ctx = make_fake_context(make_options('abc'), **ctx_kwargs)
        constr = FakeConstraint()
        constr.check(['a', 'b'], ctx)
        assert Constraint.must_check_consistency(ctx) == should_check
        assert len(constr.check_consistency_calls) == int(should_check)

    def test_error_is_raised_when_using_call_the_old_way(self):
        constr = FakeConstraint()
        with pytest.raises(TypeError, match='since Cloup v0.9, calling a constraint'):
            constr(['a', 'b'])


class TestAnd:

    @mark.parametrize('b_satisfied', [False, True])
    @mark.parametrize('a_satisfied', [False, True])
    def test_check(self, a_satisfied, b_satisfied):
        ctx = make_fake_context(make_options(['arg1', 'str_opt', 'int_opt', 'flag']))
        a = FakeConstraint(satisfied=a_satisfied)
        b = FakeConstraint(satisfied=b_satisfied)
        c = a & b
        with should_raise(ConstraintViolated, when=not (a_satisfied and b_satisfied)):
            c.check(params=['arg1', 'str_opt'], ctx=ctx)

    def test_operand_merging(self):
        a, b, c, d = (FakeConstraint() for _ in range(4))
        res = (a & b) & c
        assert res.constraints == (a, b, c)
        res = (a & b) & (c & d)
        assert res.constraints == (a, b, c, d)
        res = (a & b) & (c | d)
        assert len(res.constraints) == 3


class TestOr:
    @mark.parametrize('b_satisfied', [False, True])
    @mark.parametrize('a_satisfied', [False, True])
    def test_check(self, a_satisfied, b_satisfied):
        ctx = make_fake_context(make_options(['arg1', 'str_opt', 'int_opt', 'flag']))
        a = FakeConstraint(satisfied=a_satisfied)
        b = FakeConstraint(satisfied=b_satisfied)
        c = a | b
        with should_raise(ConstraintViolated, when=not (a_satisfied or b_satisfied)):
            c.check(params=['arg1', 'str_opt'], ctx=ctx)

    def test_operands_merging(self):
        a, b, c, d = (FakeConstraint() for _ in range(4))
        res = (a | b) | c
        assert res.constraints == (a, b, c)
        res = (a | b) | (c | d)
        assert res.constraints == (a, b, c, d)
        res = (a | b) | (c & d)
        assert len(res.constraints) == 3


def test_operator_help(dummy_ctx):
    ctx = dummy_ctx
    a, b, c = RequireAtLeast(3), AcceptAtMost(10), RequireExactly(8)
    a_help, b_help, c_help = (cons.help(ctx) for cons in [a, b, c])
    assert (a | b | c).help(ctx) == f'{a_help} or {b_help} or {c_help}'
    assert (a | b & c).help(ctx) == f'{a_help} or ({b_help} and {c_help})'


class TestRequireAtLeast:
    def test_init_raises_for_invalid_n(self):
        RequireAtLeast(0)
        with pytest.raises(ValueError):
            RequireAtLeast(-1)

    def test_help(self, dummy_ctx):
        assert '3' in RequireAtLeast(3).help(dummy_ctx)

    def test_check_consistency(self):
        check_consistency = RequireAtLeast(3).check_consistency
        check_consistency(make_options('abc'))
        with pytest.raises(UnsatisfiableConstraint):
            check_consistency(make_options('ab'))

    def test_check(self, sample_cmd: Command):
        ctx = make_context(sample_cmd, 'a1 --str-opt=ciao --bool-opt=0')
        check = partial(RequireAtLeast(2).check, ctx=ctx)
        check(['str_opt', 'int_opt', 'bool_opt'])  # str-opt and bool-opt
        check(['arg1', 'int_opt', 'def1'])  # arg1 and def1
        check(['arg1', 'str_opt', 'def1'])  # arg1, str-opt and def1
        with pytest.raises(ConstraintViolated):
            check(['str_opt', 'arg2', 'flag'])  # only str-opt is set


class TestAcceptAtMost:
    def test_init_raises_for_invalid_n(self):
        AcceptAtMost(0)
        with pytest.raises(ValueError):
            AcceptAtMost(-1)

    def test_help(self, dummy_ctx):
        assert '3' in AcceptAtMost(3).help(dummy_ctx)

    def test_check_consistency(self):
        check_consistency = AcceptAtMost(2).check_consistency
        check_consistency(make_options('abc'))
        with pytest.raises(UnsatisfiableConstraint):
            check_consistency(make_options('abc', required=True))

    def test_check(self, sample_cmd: Command):
        ctx = make_context(sample_cmd, 'a1 --str-opt=ciao --bool-opt=0')
        check = partial(AcceptAtMost(2).check, ctx=ctx)
        check(['str_opt', 'int_opt', 'bool_opt'])  # str-opt and bool-opt
        check(['arg1', 'int_opt', 'flag'])  # arg1
        with pytest.raises(ConstraintViolated):
            check(['arg1', 'str_opt', 'def1'])  # arg1, str-opt, def1


class TestRequireExactly:
    def test_init_raises_for_invalid_n(self):
        with pytest.raises(ValueError):
            RequireExactly(0)
        with pytest.raises(ValueError):
            RequireExactly(-1)

    def test_help(self, dummy_ctx):
        assert '3' in RequireExactly(3).help(dummy_ctx)

    def test_check_consistency(self):
        check_consistency = RequireExactly(3).check_consistency
        check_consistency(make_options('abcd'))
        with pytest.raises(UnsatisfiableConstraint):
            check_consistency(make_options('ab'))
        with pytest.raises(UnsatisfiableConstraint):
            check_consistency(make_options('abcde', required=True))

    def test_check(self, sample_cmd: Command):
        ctx = make_context(sample_cmd, 'a1 --str-opt=ciao --bool-opt=0')
        check = partial(RequireExactly(2).check, ctx=ctx)
        check(['str_opt', 'int_opt', 'bool_opt'])  # str-opt and bool-opt
        check(['arg1', 'int_opt', 'bool_opt'])  # arg1 and bool-opt
        with pytest.raises(ConstraintViolated):
            check(['arg1', 'str_opt', 'def1'])  # arg1, str-opt, def1
        with pytest.raises(ConstraintViolated):
            check(['arg1', 'int_opt', 'flag'])  # arg1


class TestAcceptBetween:
    def test_init_raises_for_invalid_n(self):
        AcceptBetween(0, 10)
        AcceptBetween(0, 1)
        with pytest.raises(ValueError):
            AcceptBetween(-1, 2)
        with pytest.raises(ValueError):
            AcceptBetween(2, 2)
        with pytest.raises(ValueError):
            AcceptBetween(3, 2)

    def test_help(self, dummy_ctx):
        help = AcceptBetween(3, 5).help(dummy_ctx)
        assert help == 'at least 3 required, at most 5 accepted'

    def test_check_consistency(self):
        check_consistency = AcceptBetween(2, 4).check_consistency
        check_consistency(make_options('abcd'))
        with pytest.raises(UnsatisfiableConstraint):
            check_consistency(make_options('a'))  # too little params
        with pytest.raises(UnsatisfiableConstraint):
            check_consistency(make_options('abcde', required=True))  # too many required

    def test_check(self, sample_cmd: Command):
        ctx = make_context(sample_cmd, 'a1 --str-opt=ciao --bool-opt=0 --flag --mul1=4')
        check = partial(AcceptBetween(2, 4).check, ctx=ctx)
        check(['str_opt', 'int_opt', 'bool_opt'])  # str-opt and bool-opt
        check(['arg1', 'int_opt', 'flag'])  # arg1, bool-opt and flag
        check(['def1', 'int_opt', 'flag', 'mul1'])  # all
        with pytest.raises(ConstraintViolated):
            check(['arg2', 'int_opt', 'def1'])  # only def1
        with pytest.raises(ConstraintViolated):
            check(['arg1', 'def1', 'def2', 'str_opt', 'flag'])  # all


class TestRequiredAll:
    def test_help(self, dummy_ctx):
        assert 'all required' in require_all.help(dummy_ctx)

    def test_check_consistency(self):
        require_all.check_consistency(make_options('abc'))

    def test_check(self, sample_cmd: Command):
        ctx = make_context(sample_cmd, 'arg1 --str-opt=0 --bool-opt=0')
        check = partial(require_all.check, ctx=ctx)
        check(['arg1'])
        check(['str_opt'])
        check(['arg1', 'str_opt'])
        check(['arg1', 'str_opt', 'bool_opt'])
        check(['arg1', 'str_opt', 'bool_opt', 'def1'])
        with pytest.raises(ConstraintViolated):
            check(['arg2'])
        with pytest.raises(ConstraintViolated):
            check(['arg1', 'arg2'])
        with pytest.raises(ConstraintViolated):
            check(['arg1', 'def1', 'int_opt'])


class TestRephraser:
    def test_init_raises_if_neither_help_nor_error_is_provided(self):
        with pytest.raises(ValueError):
            Rephraser(FakeConstraint())

    def test_help_override_with_string(self, dummy_ctx):
        wrapped = FakeConstraint()
        rephrased = Rephraser(wrapped, help='rephrased help')
        assert rephrased.help(dummy_ctx) == 'rephrased help'

    def test_help_override_with_function(self, dummy_ctx):
        wrapped = FakeConstraint()
        get_help = Mock(return_value='rephrased help')
        rephrased = Rephraser(wrapped, help=get_help)
        assert rephrased.help(dummy_ctx) == 'rephrased help'
        get_help.assert_called_once_with(dummy_ctx, wrapped)

    def test_error_is_overridden_passing_string(self):
        fake_ctx = make_fake_context(make_options('abcd'))
        wrapped = FakeConstraint(satisfied=False, error='__error__')
        rephrased = Rephraser(wrapped, error=f'error:\n{ErrorFmt.param_list}')
        with pytest.raises(ConstraintViolated) as exc_info:
            rephrased.check(['a', 'b'], ctx=fake_ctx)
        assert exc_info.value.message == 'error:\n  --a\n  --b\n'

    def test_error_template_key(self):
        fake_ctx = make_fake_context(make_options('abcd'))
        wrapped = FakeConstraint(satisfied=False, error='__error__')
        rephrased = Rephraser(wrapped, error=f'{ErrorFmt.error}\nExtra info here.')
        with pytest.raises(ConstraintViolated) as exc_info:
            rephrased.check(['a', 'b'], ctx=fake_ctx)
        assert str(exc_info.value) == '__error__\nExtra info here.'

    def test_error_is_overridden_passing_function(self):
        params = make_options('abc')
        fake_ctx = make_fake_context(params)
        wrapped = FakeConstraint(satisfied=False)
        error_rephraser_mock = Mock(return_value='rephrased error')
        rephrased = Rephraser(wrapped, error=error_rephraser_mock)
        with pytest.raises(ConstraintViolated, match='rephrased error'):
            rephrased.check(params, ctx=fake_ctx)

        # Check the function is called with a single argument of type ConstraintViolated
        error_rephraser_mock.assert_called_once()
        args = error_rephraser_mock.call_args[0]
        assert len(args) == 1
        error = args[0]
        assert isinstance(error, ConstraintViolated)
        # Check the error has all fields set
        assert isinstance(error.ctx, Context)
        assert isinstance(error.constraint, Constraint)
        assert len(error.params) == 3

    def test_check_consistency_raises_if_wrapped_constraint_raises(self):
        constraint = FakeConstraint(consistent=True)
        rephraser = Rephraser(constraint, help='help')
        params = make_options('abc')
        rephraser.check_consistency(params)

    def test_check_consistency_doesnt_raise_if_wrapped_constraint_doesnt_raise(self):
        constraint = FakeConstraint(consistent=False)
        rephraser = Rephraser(constraint, help='help')
        params = make_options('abc')
        with pytest.raises(UnsatisfiableConstraint):
            rephraser.check_consistency(params)



from unittest.mock import Mock

import pytest

import cloup
from cloup import Context
from tests.util import parametrize


@parametrize(
    ['ctx_arg_name', 'formatter_arg_name'],
    pytest.param('terminal_width', 'width', id='width'),
    pytest.param('max_content_width', 'max_width', id='max_width'),
)
@parametrize(
    ['ctx_arg_value', 'formatter_arg_value', 'should_warn'],
    pytest.param(80, None, False, id='only_ctx'),
    pytest.param(None, 90, False, id='only_formatter'),
    pytest.param(80, 90, True, id='both'),
)
def test_warning_is_raised_iff_arg_is_provided_both_as_context_and_formatter_arg(
    ctx_arg_name, formatter_arg_name, ctx_arg_value, formatter_arg_value, should_warn
):
    kwargs = {
        ctx_arg_name: ctx_arg_value,
        "formatter_settings": {formatter_arg_name: formatter_arg_value}
    }

    with pytest.warns(None) as warns:
        Context(command=Mock(), **kwargs)
    if should_warn:
        assert len(warns) == 1
        warning = warns[0]
        assert warning.category is UserWarning
        assert 'You provided both' in str(warning.message)
    else:
        assert len(warns) == 0


@parametrize(
    ['ctx_arg_name', 'formatter_arg_name'],
    pytest.param('terminal_width', 'width', id='width'),
    pytest.param('max_content_width', 'max_width', id='max_width'),
)
def test_warning_suppression(ctx_arg_name, formatter_arg_name):
    kwargs = {
        ctx_arg_name: 80, "formatter_settings": {formatter_arg_name: 90}
    }
    cloup.warnings.formatter_settings_conflict = False
    with pytest.warns(None) as warns:
        Context(command=Mock(), **kwargs)
    cloup.warnings.formatter_settings_conflict = True
    assert len(warns) == 0


def test_context_settings_creation():
    assert Context.settings() == {}
    assert Context.settings(
        resilient_parsing=False, align_sections=True, formatter_settings={'width': 80}
    ) == dict(
        resilient_parsing=False, align_sections=True, formatter_settings={'width': 80}
    )



from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    import cloup

FORMATTER_TYPE_ERROR = """
since cloup v0.8.0, this class relies on cloup.HelpFormatter to align help
sections. So, you need to make sure your command class uses cloup.HelpFormatter
as formatter class.

If you have your own custom HelpFormatter, know that cloup.HelpFormatter is
more easily customizable then Click's one, so consider extending it instead
of extending click.HelpFormatter.
"""


def ensure_is_cloup_formatter(formatter: click.HelpFormatter) -> 'cloup.HelpFormatter':
    from cloup import HelpFormatter
    if isinstance(formatter, HelpFormatter):
        return formatter
    raise TypeError(FORMATTER_TYPE_ERROR)


def unstyled_len(string: str) -> int:
    return len(click.unstyle(string))



from typing import (
    Callable, Iterable, List, NamedTuple, Optional, Sequence,
    TYPE_CHECKING, Tuple, Union,
)

from click import Context, Parameter

from ._core import Constraint
from .common import join_param_labels
from .._util import first_bool
from ..typing import Decorator, F

if TYPE_CHECKING:
    from .._option_groups import OptionGroup


class BoundConstraintSpec(NamedTuple):
    """A NamedTuple storing a ``Constraint`` and the **names of the parameters**
    it has to check."""
    constraint: Constraint
    param_names: Union[Sequence[str]]

    def resolve_params(self, cmd: 'ConstraintMixin') -> 'BoundConstraint':
        return BoundConstraint(
            self.constraint,
            cmd.get_params_by_name(self.param_names)
        )


def _constraint_memo(
    f, constr: Union[BoundConstraintSpec, BoundConstraintSpec]
) -> None:
    if not hasattr(f, '__constraints'):
        f.__constraints = []
    f.__constraints.append(constr)


def constraint(constr: Constraint, params: Iterable[str]) -> Callable[[F], F]:
    """Registers a constraint on a list of parameters specified by (destination) name
    (e.g. the default name of ``--input-file`` is ``input_file``)."""
    spec = BoundConstraintSpec(constr, tuple(params))

    def decorator(f):
        _constraint_memo(f, spec)
        return f

    return decorator


def constrained_params(
    constr: Constraint,
    *param_adders: Decorator,
) -> Callable[[F], F]:
    """
    Returns a decorator that adds the given parameters and applies a constraint
    to them. Equivalent to::

        @param_adders[0]
        ...
        @param_adders[-1]
        @constraint(constr, <param names>)

    This decorator saves you to manually (re)type the parameter names.
    It can also be used inside ``@option_group``.

    Instead of using this decorator, you can also call the constraint itself::

        @constr(*param_adders)

    but remember that:

    - Python 3.9 is the first that allows arbitrary expressions on the right of ``@``;
    - using a long conditional/composite constraint as decorator may be less
      readable.

    In these cases, you may consider use ``@constraint_params``.

    .. versionadded:: 0.9.0

    :param constr: an instance of :class:`Constraint`
    :param param_adders:
        function decorators, each attaching a single parameter to the decorated
        function.
    """

    def decorator(f):
        reversed_params = []
        for add_param in reversed(param_adders):
            add_param(f)
            param = f.__click_params__[-1]
            reversed_params.append(param)
        bound_constr = BoundConstraint(constr, tuple(reversed_params[::-1]))
        _constraint_memo(f, bound_constr)
        return f

    return decorator


class BoundConstraint(NamedTuple):
    """Internal utility ``NamedTuple`` that represents a ``Constraint``
    bound to a collection of ``Parameter`` instances.
    Note: this is not a subclass of Constraint."""

    constraint: Constraint
    params: Sequence[Parameter]

    def check_consistency(self):
        self.constraint.check_consistency(self.params)

    def check_values(self, ctx: Context):
        self.constraint.check_values(self.params, ctx)

    def get_help_record(self, ctx: Context) -> Optional[Tuple[str, str]]:
        constr_help = self.constraint.help(ctx)
        if not constr_help:
            return None
        param_list = '{%s}' % join_param_labels(self.params)
        return param_list, constr_help


class ConstraintMixin:
    """Provides support to constraints."""

    def __init__(
        self, *args,
        constraints: Sequence[Union[BoundConstraintSpec, BoundConstraint]] = (),
        show_constraints: Optional[bool] = None,
        **kwargs
    ):
        """
        :param constraints:
            sequence of constraints bound to specific groups of parameters.
            Note that constraints applied to option groups are collected from
            the option groups themselves, so they don't need to be included in
            this argument.
        :param show_constraints:
            whether to include a "Constraint" section in the command help. This
            is also available as a context setting having a lower priority than
            this attribute.
        :param args:
            positional arguments forwarded to the next class in the MRO
        :param kwargs:
            keyword arguments forwarded to the next class in the MRO
        """
        super().__init__(*args, **kwargs)  # type: ignore
        self.show_constraints = show_constraints

        # This allows constraints to efficiently access parameters by name
        self._params_by_name = {param.name: param for param in self.params}  # type:ignore

        # Collect constraints applied to option groups and bind them to the
        # corresponding Option instances
        option_groups: Tuple[OptionGroup, ...] = getattr(self, 'option_groups', tuple())
        self.optgroup_constraints = tuple(
            BoundConstraint(grp.constraint, grp.options)
            for grp in option_groups
            if grp.constraint is not None
        )
        """Constraints applied to ``OptionGroup`` instances."""

        # Bind constraints defined via @constraint to Parameter instances
        self.param_constraints: Tuple[BoundConstraint, ...] = tuple(
            (
                constr if isinstance(constr, BoundConstraint)
                else constr.resolve_params(self)
            )
            for constr in constraints
        )
        """Constraints registered using ``@constraint`` (or equivalent method)."""

        self.all_constraints = self.optgroup_constraints + self.param_constraints
        """All constraints applied to parameter/option groups of this command."""

    def parse_args(self, ctx: Context, args: List[str]) -> List[str]:
        # Check constraints consistency *before* parsing
        if not ctx.resilient_parsing and Constraint.must_check_consistency(ctx):
            for constr in self.all_constraints:
                constr.check_consistency()

        args = super().parse_args(ctx, args)  # type: ignore

        # Check parameters satisfy all constraints
        if not ctx.resilient_parsing:
            for constr in self.all_constraints:
                constr.check_values(ctx)
        return args

    def get_param_by_name(self, name: str) -> Parameter:
        try:
            return self._params_by_name[name]
        except KeyError:
            raise KeyError(f"there's no CLI parameter named '{name}'")

    def get_params_by_name(self, names: Iterable[str]) -> Sequence[Parameter]:
        return tuple(self.get_param_by_name(name) for name in names)

    def format_constraints(self, ctx, formatter) -> None:
        records_gen = (constr.get_help_record(ctx) for constr in self.param_constraints)
        records = [rec for rec in records_gen if rec is not None]
        if records:
            with formatter.section('Constraints'):
                formatter.write_dl(records)

    def must_show_constraints(self, ctx: Context) -> bool:
        # By default, don't show constraints
        return first_bool(
            self.show_constraints,
            getattr(ctx, "show_constraints", None),
            False,
        )


def ensure_constraints_support(command) -> ConstraintMixin:
    if isinstance(command, ConstraintMixin):
        return command
    raise TypeError(
        'a Command must inherits from ConstraintMixin to support constraints')



"""
This modules contains Cloup command classes and decorators.

Note that Cloup commands *are* Click commands. Apart from supporting more
features, Cloup command decorators have detailed type hints and are generics so
that type checkers can precisely infer the type of the returned command based on
the ``cls`` argument.

Why did you overload all decorators?
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
I wanted that the return type of decorators depended from the ``cls`` argument
but MyPy doesn't allow to set a default value on a generic argument, see:
https://github.com/python/mypy/issues/3737.
So I had to resort to a workaround using @overload which makes things more
verbose. The ``@overload`` is on the ``cls`` argument:

- in one signature, ``cls`` has type ``None`` and it's set to ``None``; in this
  case the type of the instantiated command is ``cloup.Command`` for ``@command``
  and ``cloup.Group`` for ``@group``
- in the other signature, there's ``cls: ClickCommand`` without a default, where
  ``ClickCommand`` is a type variable.

When and if the MyPy issue is resolved, the overloads will be removed.
"""
import inspect
from typing import (
    Any, Callable, Dict, Iterable, List, NamedTuple, Optional, Tuple,
    Type, TypeVar, cast, overload,
)

import click
from click import HelpFormatter

from ._context import Context
from ._option_groups import OptionGroupMixin
from ._sections import Section, SectionMixin
from ._util import first_bool, reindent, click_version_ge_8_1
from .constraints import ConstraintMixin

ClickCommand = TypeVar('ClickCommand', bound=click.Command)
ClickGroup = TypeVar('ClickGroup', bound=click.Group)


class Command(ConstraintMixin, OptionGroupMixin, click.Command):
    """A ``click.Command`` supporting option groups and constraints.

    Refer to superclasses for the documentation of all accepted parameters:

    - :class:`ConstraintMixin`
    - :class:`OptionGroupMixin`
    - :class:`click.Command`

    Besides other things, this class also:

    * back-ports a feature from Click v8.0, i.e. the ``context_class``
      class attribute, which is set to ``cloup.Context``.

    * adds a ``formatter_settings`` instance attribute.

    Refer to :class:`click.Command` for the documentation of all parameters.

    .. versionadded:: 0.8.0
    """
    context_class: Type[Context] = Context

    def __init__(
        self, *args,
        aliases: Optional[Iterable[str]] = None,
        formatter_settings: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        #: HelpFormatter options that are merged with ``Context.formatter_settings``
        #: (eventually overriding some values).
        self.aliases: List[str] = [] if aliases is None else list(aliases)
        self.formatter_settings: Dict[str, Any] = (
            {} if formatter_settings is None else formatter_settings)

    def make_context(self, info_name, args, parent=None, **extra) -> Context:
        for key, value in self.context_settings.items():
            if key not in extra:
                extra[key] = value

        ctx = self.context_class(self, info_name=info_name, parent=parent, **extra)

        with ctx.scope(cleanup=False):
            self.parse_args(ctx, args)
        return ctx

    def get_normalized_epilog(self) -> str:
        if self.epilog and click_version_ge_8_1:
            return inspect.cleandoc(self.epilog)
        return self.epilog or ""

    # Differently from Click, this doesn't indent the epilog.
    def format_epilog(self, ctx, formatter):
        if self.epilog:
            epilog = self.get_normalized_epilog()
            formatter.write_paragraph()
            formatter.write_epilog(epilog)

    def format_help_text(self, ctx, formatter):
        formatter.write_command_help_text(self)

    def format_aliases(self, ctx, formatter):
        if not self.aliases:
            return
        formatter.write_aliases(self.aliases)

    def format_help(self, ctx: click.Context, formatter: HelpFormatter) -> None:
        self.format_usage(ctx, formatter)
        self.format_aliases(ctx, formatter)
        self.format_help_text(ctx, formatter)
        self.format_params(ctx, formatter)
        if self.must_show_constraints(ctx):
            self.format_constraints(ctx, formatter)
        if isinstance(self, click.MultiCommand):
            self.format_commands(ctx, formatter)
        self.format_epilog(ctx, formatter)


class Group(SectionMixin, Command, click.Group):
    """
    A ``click.Group`` that allows to organize its subcommands in multiple help
    sections and whose subcommands are, by default, of type :class:`cloup.Command`.

    Refer to superclasses for the documentation of all accepted parameters:

    - :class:`SectionMixin`
    - :class:`Command`
    - :class:`click.Group`

    Apart from superclasses arguments, the following is the only additional parameter:

    ``show_subcommand_aliases``: ``Optional[bool] = None``
        whether to show subcommand aliases; aliases are shown by default and
        can be disabled using this argument or the homonym context setting.

    .. versionchanged:: 0.14.0
        this class now supports option groups and constraints.

    .. versionadded:: 0.10.0
        the "command aliases" feature, including the ``show_subcommand_aliases``
        parameter/attribute.

    .. versionchanged:: 0.8.0
        this class now inherits from :class:`cloup.BaseCommand`.
    """
    SHOW_SUBCOMMAND_ALIASES: bool = False

    def __init__(self, *args, show_subcommand_aliases: Optional[bool] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.show_subcommand_aliases = show_subcommand_aliases
        """Whether to show subcommand aliases. This """

        self.alias2name: Dict[str, str] = {}
        """Dictionary mapping each alias to a command name."""

    def add_command(
        self, cmd: click.Command,
        name: Optional[str] = None,
        section: Optional[Section] = None,
        fallback_to_default_section: bool = True,
    ) -> None:
        super().add_command(cmd, name, section, fallback_to_default_section)
        name = cast(str, cmd.name) if name is None else name
        aliases = getattr(cmd, 'aliases', [])
        for alias in aliases:
            self.alias2name[alias] = name

    def resolve_command_name(self, ctx: click.Context, name: str) -> Optional[str]:
        """Maps a string supposed to be a command name or an alias to a normalized
        command name. If no match is found, it returns ``None``."""
        if ctx.token_normalize_func:
            name = ctx.token_normalize_func(name)
        if name in self.commands:
            return name
        return self.alias2name.get(name)

    def resolve_command(
        self, ctx: click.Context, args: List[str]
    ) -> Tuple[Optional[str], Optional[click.Command], List[str]]:
        normalized_name = self.resolve_command_name(ctx, args[0])
        if normalized_name:
            # Replacing this string ensures that super().resolve_command() returns a
            # normalized command name rather than an alias. The technique described in
            # Click's docs doesn't work if the subcommand is added using Group.group
            # passing the "name" argument.
            args[0] = normalized_name
        try:
            return super().resolve_command(ctx, args)
        except click.UsageError as error:
            new_error = self.handle_bad_command_name(
                bad_name=args[0],
                valid_names=[*self.commands, *self.alias2name],
                error=error
            )
            raise new_error

    def handle_bad_command_name(
        self, bad_name: str, valid_names: List[str], error: click.UsageError
    ) -> click.UsageError:
        """This method is called when a command name cannot be resolved.
        Useful to implement the "Did you mean <x>?" feature.

        :param bad_name: the command name that could not be resolved.
        :param valid_names: the list of valid command names, including aliases.
        :param error: the original error coming from Click.
        :return: the original error or a new one.
        """
        import difflib
        matches = difflib.get_close_matches(bad_name, valid_names)
        if not matches:
            return error
        elif len(matches) == 1:
            extra_msg = f"Did you mean '{matches[0]}'?"
        else:
            matches_list = "\n".join("   " + match for match in matches)
            extra_msg = 'Did you mean one of these?\n' + matches_list

        error_msg = str(error) + " " + extra_msg
        return click.exceptions.UsageError(error_msg, error.ctx)

    def must_show_subcommand_aliases(self, ctx: click.Context) -> bool:
        return first_bool(
            self.show_subcommand_aliases,
            getattr(ctx, 'show_subcommand_aliases', None),
            Group.SHOW_SUBCOMMAND_ALIASES,
        )

    def format_subcommand_name(
        self, ctx: click.Context, name: str, cmd: click.Command
    ) -> str:
        aliases = getattr(cmd, 'aliases', None)
        if aliases and self.must_show_subcommand_aliases(ctx):
            alias_list = ', '.join(aliases)
            return f"{name} ({alias_list})"
        return name

    # MyPy complains because "Signature of "group" incompatible with supertype".
    # The supertype signature is (*args, **kwargs), which is compatible with
    # this provided that you pass all arguments (expect "name") as keyword arg.
    @overload  # type: ignore
    def command(  # Why overloading? Refer to module docstring.
        self, name: Optional[str] = None,
        *,
        aliases: Optional[Iterable[str]] = None,
        cls: None = None,  # Command is cloup.Command
        section: Optional[Section] = None,
        context_settings: Optional[Dict[str, Any]] = None,
        formatter_settings: Optional[Dict[str, Any]] = None,
        help: Optional[str] = None,
        epilog: Optional[str] = None,
        short_help: Optional[str] = None,
        options_metavar: Optional[str] = "[OPTIONS]",
        add_help_option: bool = True,
        no_args_is_help: bool = False,
        hidden: bool = False,
        deprecated: bool = False,
        align_option_groups: Optional[bool] = None,
        show_constraints: Optional[bool] = None,
        params: Optional[List[click.Parameter]] = None,
    ) -> Callable[[Callable], Command]:
        ...

    @overload  # type: ignore
    def command(  # Why overloading? Refer to module docstring.
        self, name: Optional[str] = None,
        *,
        aliases: Optional[Iterable[str]] = None,
        cls: Type[ClickCommand],
        section: Optional[Section] = None,
        context_settings: Optional[Dict[str, Any]] = None,
        help: Optional[str] = None,
        epilog: Optional[str] = None,
        short_help: Optional[str] = None,
        options_metavar: Optional[str] = "[OPTIONS]",
        add_help_option: bool = True,
        no_args_is_help: bool = False,
        hidden: bool = False,
        deprecated: bool = False,
        params: Optional[List[click.Parameter]] = None,
        **kwargs,
    ) -> Callable[[Callable], ClickCommand]:
        ...

    def command(self, name=None, *, cls=None, section=None, **kwargs):
        """Returns a decorator that creates a new subcommand of this ``Group``
        using the decorated function as callback.

        It takes the same arguments of :func:`command` plus:

        ``section``: ``Optional[Section]``
            if provided, put the subcommand in this section.

        .. versionchanged:: 0.10.0
            all arguments but ``name`` are now keyword-only.
        """

        def decorator(f):
            cmd = command(name=name, cls=cls, **kwargs)(f)
            self.add_command(cmd, section=section)
            return cmd

        return decorator

    # MyPy complains because "Signature of "group" incompatible with supertype".
    # The supertype signature is (*args, **kwargs), which is compatible with
    # this provided that you pass all arguments (expect "name") as keyword arg.
    @overload  # type: ignore
    def group(  # Why overloading? Refer to module docstring.
        self, name: Optional[str] = None,
        *,
        aliases: Optional[Iterable[str]] = None,
        cls: None = None,  # cls not provided
        section: Optional[Section] = None,
        sections: Iterable[Section] = (),
        align_sections: Optional[bool] = None,
        invoke_without_command: bool = False,
        no_args_is_help: bool = False,
        context_settings: Optional[Dict[str, Any]] = None,
        formatter_settings: Dict[str, Any] = {},
        help: Optional[str] = None,
        epilog: Optional[str] = None,
        short_help: Optional[str] = None,
        options_metavar: Optional[str] = "[OPTIONS]",
        subcommand_metavar: Optional[str] = None,
        add_help_option: bool = True,
        chain: bool = False,
        hidden: bool = False,
        deprecated: bool = False,
    ) -> Callable[[Callable], 'Group']:
        ...

    @overload
    def group(  # Why overloading? Refer to module docstring.
        self, name: Optional[str] = None, *,
        aliases: Optional[Iterable[str]] = None,
        cls: Type[ClickGroup],
        section: Optional[Section] = None,
        invoke_without_command: bool = False,
        no_args_is_help: bool = False,
        context_settings: Optional[Dict[str, Any]] = None,
        help: Optional[str] = None,
        epilog: Optional[str] = None,
        short_help: Optional[str] = None,
        options_metavar: Optional[str] = "[OPTIONS]",
        subcommand_metavar: Optional[str] = None,
        add_help_option: bool = True,
        chain: bool = False,
        hidden: bool = False,
        deprecated: bool = False,
        params: Optional[List[click.Parameter]] = None,
        **kwargs
    ) -> Callable[[Callable], ClickGroup]:
        ...

    def group(self, name=None, *, cls=None, section=None, **kwargs):
        """Returns a decorator that creates a new subcommand of this ``Group``
        using the decorated function as callback.

        It takes the same argument of :func:`group` plus:

        ``section``: ``Optional[Section]``
            if provided, put the subcommand in this section.

        .. versionchanged:: 0.10.0
            all arguments but ``name`` are now keyword-only.
        """

        def decorator(f):
            cmd = group(name=name, cls=cls, **kwargs)(f)
            self.add_command(cmd, section=section)
            return cmd

        return decorator


# Why overloading? Refer to module docstring.
@overload  # In this overload: "cls: None = None"
def command(
    name: Optional[str] = None,
    *,
    aliases: Optional[Iterable[str]] = None,
    cls: None = None,
    context_settings: Optional[Dict[str, Any]] = None,
    formatter_settings: Optional[Dict[str, Any]] = None,
    help: Optional[str] = None,
    short_help: Optional[str] = None,
    epilog: Optional[str] = None,
    options_metavar: Optional[str] = "[OPTIONS]",
    add_help_option: bool = True,
    no_args_is_help: bool = False,
    hidden: bool = False,
    deprecated: bool = False,
    align_option_groups: Optional[bool] = None,
    show_constraints: Optional[bool] = None,
    params: Optional[List[click.Parameter]] = None,
) -> Callable[[Callable], Command]:
    ...


@overload
def command(  # In this overload: "cls: ClickCommand"
    name: Optional[str] = None,
    *,
    aliases: Optional[Iterable[str]] = None,
    cls: Type[ClickCommand],
    context_settings: Optional[Dict[str, Any]] = None,
    help: Optional[str] = None,
    short_help: Optional[str] = None,
    epilog: Optional[str] = None,
    options_metavar: Optional[str] = "[OPTIONS]",
    add_help_option: bool = True,
    no_args_is_help: bool = False,
    hidden: bool = False,
    deprecated: bool = False,
    params: Optional[List[click.Parameter]] = None,
    **kwargs
) -> Callable[[Callable], ClickCommand]:
    ...


# noinspection PyIncorrectDocstring
def command(name=None, *, aliases=None, cls=None, **kwargs):
    """
    Returns a decorator that creates a new command using the decorated function
    as callback.

    The only differences with respect to ``click.command`` are:

    - the default command class is :class:`cloup.Command`
    - supports constraints, provided that ``cls`` inherits from ``ConstraintMixin``
      like ``cloup.Command`` (the default)
    - this function has detailed type hints and uses generics for the ``cls``
      argument and return type.

    Note that the following arguments are about Cloup-specific features and are
    not supported by all ``click.Command``, so if you provide a custom ``cls``
    make sure you don't set these:

    - ``formatter_settings``
    - ``align_option_groups`` (``cls`` needs to inherit from ``OptionGroupMixin``)
    - ``show_constraints`` (``cls`` needs to inherit ``ConstraintMixin``).

    .. versionchanged:: 0.10.0
        this function is now generic: the return type depends on what you provide
        as ``cls`` argument.

    .. versionchanged:: 0.9.0
        all arguments but ``name`` are now keyword-only arguments.

    :param name:
        the name of the command to use unless a group overrides it.
    :param aliases:
        alternative names for this command. If ``cls`` is not a Cloup command class,
        aliases will be stored in the instantiated command by monkey-patching
        and aliases won't be documented in the help page of the command.
    :param cls:
        the command class to instantiate.
    :param context_settings:
        an optional dictionary with defaults that are passed to the context object.
    :param formatter_settings:
        arguments for the formatter; you can use :meth:`HelpFormatter.settings`
        to build this dictionary.
    :param help:
        the help string to use for this command.
    :param epilog:
        like the help string but it's printed at the end of the help page after
        everything else.
    :param short_help:
        the short help to use for this command.  This is shown on the command
        listing of the parent command.
    :param options_metavar:
        metavar for options shown in the command's usage string.
    :param add_help_option:
        by default each command registers a ``--help`` option.
        This can be disabled by this parameter.
    :param no_args_is_help:
        this controls what happens if no arguments are provided. This option is
        disabled by default. If enabled this will add ``--help`` as argument if
        no arguments are passed
    :param hidden:
        hide this command from help outputs.
    :param deprecated:
        issues a message indicating that the command is deprecated.
    :param align_option_groups:
        whether to align the columns of all option groups' help sections.
        This is also available as a context setting having a lower priority
        than this attribute. Given that this setting should be consistent
        across all you commands, you should probably use the context
        setting only.
    :param show_constraints:
        whether to include a "Constraint" section in the command help. This
        is also available as a context setting having a lower priority than
        this attribute.
    :param params:
        **(click >= 8.1.0)** a list of parameters (:class:`Argument` and
        :class:`Option` instances). Params added with ``@option`` and ``@argument``
        are appended to the end of the list if given.
    :param kwargs:
        any other argument accepted by the instantiated command class (``cls``).
    """

    def decorator(f):
        if hasattr(f, '__constraints'):
            if cls and not issubclass(cls, ConstraintMixin):
                raise TypeError(
                    f"a Command must inherit from cloup.ConstraintMixin to support "
                    f"constraints; {cls} doesn't")
            constraints = tuple(reversed(f.__constraints))
            del f.__constraints
            kwargs['constraints'] = constraints

        cmd_cls = cls if cls is not None else Command
        try:
            cmd = click.command(name, cls=cmd_cls, **kwargs)(f)
            if aliases:
                cmd.aliases = list(aliases)
            return cmd
        except TypeError as error:
            raise _process_unexpected_kwarg_error(error, _ARGS_INFO, cls)

    return decorator


@overload  # Why overloading? Refer to module docstring.
def group(
    name: Optional[str] = None,
    *,
    aliases: Optional[Iterable[str]] = None,
    cls: None = None,
    sections: Iterable[Section] = (),
    align_sections: Optional[bool] = None,
    invoke_without_command: bool = False,
    no_args_is_help: bool = False,
    context_settings: Optional[Dict[str, Any]] = None,
    formatter_settings: Dict[str, Any] = {},
    help: Optional[str] = None,
    short_help: Optional[str] = None,
    epilog: Optional[str] = None,
    options_metavar: Optional[str] = "[OPTIONS]",
    subcommand_metavar: Optional[str] = None,
    add_help_option: bool = True,
    chain: bool = False,
    hidden: bool = False,
    deprecated: bool = False,
    params: Optional[List[click.Parameter]] = None,
) -> Callable[[Callable], Group]:
    ...


@overload
def group(
    name: Optional[str] = None,
    *,
    aliases: Optional[Iterable[str]] = None,
    cls: Type[ClickGroup],
    invoke_without_command: bool = False,
    no_args_is_help: bool = False,
    context_settings: Optional[Dict[str, Any]] = None,
    help: Optional[str] = None,
    short_help: Optional[str] = None,
    epilog: Optional[str] = None,
    options_metavar: Optional[str] = "[OPTIONS]",
    subcommand_metavar: Optional[str] = None,
    add_help_option: bool = True,
    chain: bool = False,
    hidden: bool = False,
    deprecated: bool = False,
    params: Optional[List[click.Parameter]] = None,
    **kwargs
) -> Callable[[Callable], ClickGroup]:
    ...


def group(name=None, *, cls=None, **kwargs):
    """
    Returns a decorator that instantiates a ``Group`` (or a subclass of it)
    using the decorated function as callback.

    .. versionchanged:: 0.10.0
        the ``cls`` argument can now be any ``click.Group`` (previously had to
        be a ``cloup.Group``) and the type of the instantiated command matches
        it (previously, the type was ``cloup.Group`` even if ``cls`` was a subclass
        of it).

    .. versionchanged:: 0.9.0
        all arguments but ``name`` are now keyword-only arguments.

    :param name:
        the name of the command to use unless a group overrides it.
    :param cls:
        the ``click.Group`` (sub)class to instantiate. This is ``cloup.Group``
        by default. Note that some of the arguments are only supported by
        ``cloup.Group``.
    :param sections:
        a list of Section objects containing the subcommands of this ``Group``.
        This argument is only supported by commands inheriting from
        :class:`cloup.SectionMixin`.
    :param align_sections:
        whether to align the columns of all subcommands' help sections.
        This is also available as a context setting having a lower priority
        than this attribute. Given that this setting should be consistent
        across all you commands, you should probably use the context
        setting only.
    :param context_settings:
        an optional dictionary with defaults that are passed to the context object.
    :param formatter_settings:
        arguments for the formatter; you can use :meth:`HelpFormatter.settings`
        to build this dictionary.
    :param help:
        the help string to use for this command.
    :param short_help:
        the short help to use for this command.  This is shown on the command
        listing of the parent command.
    :param epilog:
        like the help string but it's printed at the end of the help page after
        everything else.
    :param options_metavar:
        metavar for options shown in the command's usage string.
    :param add_help_option:
        by default each command registers a ``--help`` option.
        This can be disabled by this parameter.
    :param hidden:
        hide this command from help outputs.
    :param deprecated:
        issues a message indicating that the command is deprecated.
    :param invoke_without_command:
        this controls how the multi command itself is invoked. By default it's
        only invoked if a subcommand is provided.
    :param no_args_is_help:
        this controls what happens if no arguments are provided. This option is
        enabled by default if `invoke_without_command` is disabled or disabled
        if it's enabled. If enabled this will add ``--help`` as argument if no
        arguments are passed.
    :param subcommand_metavar:
        string used in the command's usage string to indicate the subcommand place.
    :param chain:
        if this is set to `True` chaining of multiple subcommands is enabled.
        This restricts the form of commands in that they cannot have optional
        arguments but it allows multiple commands to be chained together.
    :param params:
        **(click >= 8.1.0)** a list of parameters (:class:`Argument` and
        :class:`Option` instances). Params added with ``@option`` and ``@argument``
        are appended to the end of the list if given.
    :param kwargs:
        any other argument accepted by the instantiated command class.
    """
    if cls is None:
        cls = Group
    elif not issubclass(cls, click.Group):
        raise TypeError(
            'this decorator requires cls to be a click.Group (or a subclass).')

    return command(name=name, cls=cls, **kwargs)


# Side stuff for better error messages

class _ArgInfo(NamedTuple):
    arg_name: str
    requires: Type
    supported_by: str = ""


_ARGS_INFO = {
    info.arg_name: info for info in [
        _ArgInfo('formatter_settings', Command, "both Command and Group"),
        _ArgInfo('align_option_groups', OptionGroupMixin, "both Command and Group"),
        _ArgInfo('show_constraints', ConstraintMixin, "both Command and Group"),
        _ArgInfo('align_sections', SectionMixin, "Group")
    ]
}


def _process_unexpected_kwarg_error(
    error: TypeError, args_info: Dict[str, _ArgInfo], cls: Type[click.Command]
) -> TypeError:
    """Checks if the developer tried to pass a Cloup-specific argument to a ``cls``
    that doesn't support it and if that's the case, augments the error message
    to provide useful more info about the error."""
    import re

    message = str(error)
    match = re.search('|'.join(arg_name for arg_name in args_info), message)
    if match is None:
        return error
    arg = match.group()
    info = args_info[arg]
    extra_info = reindent(f"""\n
        HINT: you set cls={cls} but this class doesn't support the argument "{arg}".
        In Cloup, this argument is supported by {info.supported_by}
        via {info.requires.__name__}.
    """, 4)
    new_message = message + '\n' + extra_info
    return TypeError(new_message)



"""
Show-casing the "Positional arguments" help section.
"""
from pprint import pprint

import cloup
from cloup import option, option_group


@cloup.command(name='cloup', show_constraints=True)
@cloup.argument('input_path', help="Input path")
@cloup.argument('out_path', help="Output path")
@option_group(
    'An option group',
    option('-o', '--one', help='a 1st cool option'),
    option('-t', '--two', help='a 2nd cool option'),
    option('--three', help='a 3rd cool option'),
)
def main(**kwargs):
    """A test program for cloup."""
    pprint(kwargs, indent=3)


if __name__ == '__main__':
    main()

"""
Usage: arguments_with_help.py [OPTIONS] INPUT_PATH OUT_PATH

  A test program for cloup.

Positional arguments:
  INPUT_PATH      Input path
  OUT_PATH        Output path

An option group:
  -o, --one TEXT  a 1st cool option
  -t, --two TEXT  a 2nd cool option
  --three TEXT    a 3rd cool option

Other options:
  --help          Show this message and exit.
"""



"""
Example of options groups, "nested style" (recommended).

NOTE: the goal of this example is to showcase Cloup's option groups and constraint API
    in a compact way; the example is neither meaningful nor represents a good usage
    of the library (in fact, you should always design your CLI so that it doesn't
    need constraints).
"""
from pprint import pprint

from click import Choice

import cloup
from cloup import option, option_group
from cloup.constraints import (
    Equal, If, RequireAtLeast, RequireExactly, constraint, mutually_exclusive,
)


@cloup.command(name='cloup', show_constraints=True)
@cloup.argument('input_path')
@cloup.argument('out_path')
@option_group(
    'First group title',
    "This is a very long description of the option group. I don't think this is "
    "needed very often; still, if you want to provide it, you can pass it as 2nd "
    "positional argument or as keyword argument 'help' after all options.",
    option('-o', '--one', help='a 1st cool option'),
    option('-t', '--two', help='a 2nd cool option'),
    option('--three', help='a 3rd cool option'),
    constraint=RequireAtLeast(1),
)
@option_group(
    'Second group name',
    option('--four', help='a 4th cool option'),
    option('--five', help='a 5th cool option'),
    option('--six', help='a 6th cool option'),
    constraint=If('three', then=RequireExactly(1)),  # conditional constraint
)
@option('--seven', help='an uncategorized option', type=Choice(['foo', 'bar']))
@option('--eight', help='second uncategorized option')
# Usage of @constraint
@constraint(mutually_exclusive, ['one', 'two'])
@constraint(
    If(Equal('one', '123'), then=RequireExactly(1)),
    ['seven', 'six']
)
def main(**kwargs):
    """A test program for cloup."""
    pprint(kwargs, indent=3)


if __name__ == '__main__':
    main()



from functools import partial

import pytest

from cloup.formatting.sep import (
    Hline, RowSepIf, count_multiline_rows, multiline_rows_are_at_least
)

# Use the same widths for both columns
cols_width = 30
col_widths = (cols_width, cols_width)
col_spacing = 2

# We'll use these in many tests (a = above_limit, b = below_limit)
above_limit = 'a' * (cols_width + 1)
below_limit = 'b' * cols_width
aa = (above_limit, above_limit)
ab = (above_limit, below_limit)
ba = (below_limit, above_limit)
bb = (below_limit, below_limit)


def test_count_multiline_rows():
    count = partial(count_multiline_rows, col_widths=col_widths)
    assert count([bb]) == 0
    assert count([ba]) == 1
    assert count([ab]) == 1
    assert count([aa]) == 1
    assert count([bb, ba, ab, aa]) == 3
    with pytest.raises(Exception):
        count([tuple('1234')])  # len(row) > len(col_widths)


class TestMultilineRowsAreAtLeast:
    @pytest.mark.parametrize('bad_value', [0, 0.0, -1, -1.4, 1.1, 11.0])
    def test_args_validation(self, bad_value):
        with pytest.raises(ValueError):
            multiline_rows_are_at_least(bad_value)

    def test_with_count(self):
        at_least_2_multiline_rows = partial(multiline_rows_are_at_least(2),
                                            col_widths=(30, 30), col_spacing=2)
        assert at_least_2_multiline_rows([ab, bb, bb, ba])
        assert at_least_2_multiline_rows([ab, bb, bb, ba, aa])
        assert not at_least_2_multiline_rows([bb, bb, bb])

    def test_with_percentage(self):
        at_least_half_multiline_rows = partial(multiline_rows_are_at_least(.5),
                                               col_widths=(30, 30), col_spacing=2)
        assert at_least_half_multiline_rows([
            bb, bb, bb,
            aa, ab, ba,
        ])
        assert at_least_half_multiline_rows([
            bb, bb, bb,
            aa, ab, ba, aa, ab
        ])
        assert not at_least_half_multiline_rows([
            bb, bb, bb, bb,
            aa, ab, ba,
        ])
        assert not at_least_half_multiline_rows([bb] * 5)


def test_value_error_if_sep_string_ends_with_newline():
    with pytest.raises(ValueError, match=r"sep must not end with '\\n'"):
        RowSepIf(multiline_rows_are_at_least(1), sep='\n')


def test_Hline():
    assert Hline.solid(5) == "─────"
    assert Hline.dashed(5) == "-----"
    assert Hline.densely_dashed(5) == "╌╌╌╌╌"
    assert Hline.dotted(5) == "┄┄┄┄┄"
    assert Hline('-.')(5) == '-.-.-'



"""
This module contains a boolean variable for each warning that may be raised
by Cloup. To suppress a warning, set the corresponding variable to ``False``.
This is an alternative to using ``warnings.filterwarnings`` that:

- IMHO, offers a better developer experience since setting a boolean is
  easier and cleaner than writing a regex that matches a warning message
- can be used by Cloup itself to skip the checks that (may) generate a warning.
"""

formatter_settings_conflict = True



import re
import sys

regex = re.compile(r'^([a-zA-Z_-]+):.*?## (.*)$')

for line in sys.stdin:
    match = regex.match(line)
    if match:
        target, help = match.groups()
        print("%-20s %s" % (target, help))



import os

import cloup


@cloup.group(
    'config', aliases=['conf', 'cfg'],
    invoke_without_command=True,
    no_args_is_help=True,
)
def config():
    """Manages Manim configuration files."""
    pass


@config.command(no_args_is_help=True)
@cloup.option(
    "-l", "--level",
    type=cloup.Choice(["user", "cwd"], case_sensitive=False), default="cwd",
    help="Specify if this config is for user or the working directory.",
)
@cloup.option("-o", "--open", "openfile", is_flag=True)
def write(level: str, openfile: bool) -> None:
    """Writes configurations."""
    pass


@config.command()
def show():
    """Shows current configuration."""
    pass


@config.command()
@cloup.option("-d", "--directory", default=os.getcwd())
def export(directory):
    pass



"""Top-level package for cloup."""
# WARNING: _version.py is generated by setuptools-scm upon package building/installation
from . import _version

__author__ = """Gianluca Gippetto"""
__email__ = 'gianluca.gippetto@gmail.com'
__version__ = _version.version
__version_tuple__ = _version.version_tuple

# flake8: noqa F401
from click import (
    # decorators
    confirmation_option,
    help_option,
    pass_context,
    pass_obj,
    password_option,
    version_option,
    # types
    BOOL,
    Choice,
    DateTime,
    File,
    FLOAT,
    FloatRange,
    INT,
    IntRange,
    ParamType,
    Path,
    STRING,
    Tuple,
    UNPROCESSED,
    UUID,
)

from . import warnings
from .styling import (
    HelpTheme,
    Style,
    Color,
)
from .formatting import (
    HelpFormatter,
    HelpSection,
)
from ._context import Context
from ._params import Argument, Option, argument, option
from ._option_groups import (
    OptionGroup,
    OptionGroupMixin,
    option_group,
)
from ._sections import (
    Section,
    SectionMixin,
)
from ._commands import (
    Command,
    Group,
    command,
    group,
)
from .constraints import (
    ConstraintMixin,
    constraint,
    constrained_params,
)
from .types import dir_path, file_path, path



__all__ = ['MISSING', 'Possibly',  'Decorator', 'F']

from enum import Enum
from typing import Callable, TypeVar, Union


# PEP-blessed solution for defining a Singleton type:
# https://www.python.org/dev/peps/pep-0484/#id30
class _Missing(Enum):
    flag = 'Missing'


MISSING = _Missing.flag
"""Singleton that works as a sentinel for a missing value.
Useful when None can't be used to play the role because it represents a valid
non-null value."""

_T = TypeVar('_T')
Possibly = Union[_Missing, _T]
"""Possibly[T] is like Optional[T] but uses MISSING for missing values."""

F = TypeVar('F', bound=Callable)
"""Type variable for a Callable."""

Decorator = Callable[[Callable], Callable]
"""Type alias for a simple function decorator."""



import pytest

from cloup._util import check_positive_int, coalesce, first_bool, make_repr


def test_make_repr():
    expected_repr = "list('arg', name='Alan', surname='Turing')"
    n = len(expected_repr)
    r = make_repr([], 'arg', name='Alan', surname='Turing',
                  _line_len=n)
    assert r == expected_repr

    r = make_repr([], 'arg', name='Alan', surname='Turing',
                  _line_len=n - 1, _indent=3)
    assert r == ("list(\n"
                 "   'arg',\n"
                 "   name='Alan',\n"
                 "   surname='Turing'\n"
                 ")")


def test_check_positive_int():
    check_positive_int(1, 'name')
    with pytest.raises(ValueError):
        check_positive_int(0, 'name')
    with pytest.raises(ValueError):
        check_positive_int(-1, 'name')
    with pytest.raises(TypeError):
        check_positive_int(None, 'name')
    with pytest.raises(TypeError):
        check_positive_int(1.23, 'name')


def test_first_bool():
    assert first_bool(0, 1, None, '', [], False, True) is False
    assert first_bool(0, 1, None, '', [], True, False) is True
    with pytest.raises(StopIteration):
        first_bool(1, 2, '', [], None)


def test_coalesce():
    assert coalesce() is None
    assert coalesce(None, None, None) is None
    for expected in [0, '', [], False, 123]:
        assert coalesce(None, None, expected, True, 12) == expected
